"""Tests for the compiler module."""

import logging
import os
from pathlib import Path

import pytest
from jinja2 import UndefinedError

from astrality.compiler import (
    compile_template,
    compile_template_to_string,
    jinja_environment,
)
from astrality.context import Context


@pytest.fixture
def test_templates_folder():
    return Path(__file__).parent / 'templates'


@pytest.fixture
def jinja_test_env(test_templates_folder):
    return jinja_environment(
        test_templates_folder,
        shell_command_working_directory=Path('~').resolve(),
    )


def test_rendering_environment_variables(jinja_test_env):
    template = jinja_test_env.get_template('env_vars')
    assert template.render() == \
        'test_value\nfallback_value\n'


def test_logging_undefined_variables(jinja_test_env, caplog):
    template = jinja_test_env.get_template('env_vars')
    template.render()
    assert (
        'astrality.compiler',
        logging.WARNING,
        'Template variable warning: env_UNDEFINED_VARIABLE is undefined',
    ) in caplog.record_tuples


def test_integer_indexed_templates(jinja_test_env):
    template = jinja_test_env.get_template('integer_indexed')
    context = Context({'section': {1: 'one', 2: 'two'}})
    assert template.render(context) == 'one\ntwo\ntwo'


def test_compilation_of_jinja_template(test_templates_folder):
    template = test_templates_folder / 'env_vars'
    target = Path('/tmp/astrality') / template.name
    compile_template(template, target, {}, Path('/'))

    with open(target) as target:
        assert target.read() == 'test_value\nfallback_value\n'


def test_run_shell_template_filter(test_templates_folder):
    shell_template_path = test_templates_folder / 'shell_filter.template'
    compiled_shell_template_path = Path(
        '/tmp/astrality',
        shell_template_path.name,
    )
    compiled_shell_template_path.touch()

    context = {}
    compile_template(
        template=shell_template_path,
        target=compiled_shell_template_path,
        context=context,
        shell_command_working_directory=Path('/'),
    )

    with open(compiled_shell_template_path) as target:
        assert target.read() \
            == 'quick\nanother_quick\nslow_but_allowed\n\nfallback'

    if compiled_shell_template_path.is_file():
        os.remove(compiled_shell_template_path)


def test_working_directory_of_shell_command_filter(test_templates_folder):
    shell_template_path = Path(
        test_templates_folder,
        'shell_filter_working_directory.template',
    )
    compiled_shell_template_path = Path(
        '/tmp/astrality',
        shell_template_path.name,
    )
    context = {}
    compile_template(
        template=shell_template_path,
        target=compiled_shell_template_path,
        context=context,
        shell_command_working_directory=Path('/'),
    )

    with open(compiled_shell_template_path) as target:
        assert target.read() == '/'


def test_environment_variable_interpolation_by_preprocessing_conf_yaml_file():
    test_conf = Path(__file__).parent / 'test_config' / 'test.yml'
    result = compile_template_to_string(
        template=test_conf,
        context={},
    )

    expected_result = '\n'.join((
        'section1:',
        '    var1: value1',
        '    var2: value1/value2',
        '',
        '',
        'section2:',
        '    # Comment',
        '    var3: value1',
        "    empty_string_var: ''",
        '',
        'section3:',
        '    env_variable: test_value, hello',
        '',
        'section4:',
        '    1: primary_value',
    ))
    assert expected_result == result


@pytest.mark.slow
def test_command_substition_by_preprocessing_yaml_file():
    test_conf = Path(__file__).parent / 'test_config' / 'commands.yml'
    result = compile_template_to_string(
        template=test_conf,
        context={},
    )

    expected_result = '\n'.join((
        'section1:',
        '    key1: test',
        '    key2: test_value',
        '    key3: test_value',
        '    key4: ',
    ))
    assert expected_result == result


def test_handling_of_undefined_context(tmpdir, caplog):
    template = Path(tmpdir) / 'template'
    template.write_text('{{ this.is.not.defined }}')

    with pytest.raises(UndefinedError):
        compile_template_to_string(
            template=template,
            context={},
        )

    assert "'this' is undefined" in caplog.record_tuples[0][2]


def test_writing_template_file_with_default_permissions(tmpdir):
    tmpdir = Path(tmpdir)
    template = tmpdir / 'template'
    template.write_text('content')
    permission = 0o751
    template.chmod(permission)
    target = tmpdir / 'target'

    compile_template(
        template=template,
        target=target,
        context={},
        shell_command_working_directory=tmpdir,
    )
    assert (target.stat().st_mode & 0o777) == permission


def test_writing_template_file_with_specific_octal_permissions(tmpdir):
    tmpdir = Path(tmpdir)
    template = tmpdir / 'template'
    template.write_text('content')
    target = tmpdir / 'target'

    permissions = '514'
    compile_template(
        template=template,
        target=target,
        context={},
        shell_command_working_directory=tmpdir,
        permissions=permissions,
    )
    assert (target.stat().st_mode & 0o777) == 0o514


def test_writing_template_file_with_specific_modal_permissions(tmpdir):
    tmpdir = Path(tmpdir)
    template = tmpdir / 'template'
    template.write_text('content')
    permission = 0o644
    template.chmod(permission)
    target = tmpdir / 'target'

    permissions = 'uo+x'
    compile_template(
        template=template,
        target=target,
        context={},
        shell_command_working_directory=tmpdir,
        permissions=permissions,
    )
    assert (target.stat().st_mode & 0o777) == 0o745


def test_writing_template_file_with_invalid_permissions(tmpdir):
    tmpdir = Path(tmpdir)
    template = tmpdir / 'template'
    template.write_text('content')
    permission = 0o732
    template.chmod(permission)
    target = tmpdir / 'target'

    permissions = 'invalid_argument'
    compile_template(
        template=template,
        target=target,
        context={},
        shell_command_working_directory=tmpdir,
        permissions=permissions,
    )
    assert (target.stat().st_mode & 0o777) == 0o732



"""Custom exception definitions."""


class AstralityConfigurationError(BaseException):
    """Exception for when astrality configuration is invalid."""


class NonExistentEnabledModule(AstralityConfigurationError):
    """Exception for when an enabled module is not found."""


class MisconfiguredConfigurationFile(AstralityConfigurationError):
    """Exception for when astrality configuration has semantic errors."""


class GithubModuleError(AstralityConfigurationError):
    """Exception for when GitHub module could not be sourced."""



"""Module for directory modification watching."""

from pathlib import Path
import logging
from sys import platform
from typing import Callable

from watchdog.events import FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer


class DirectoryWatcher:
    """A directory watcher class."""

    def __init__(
        self,
        directory: Path,
        on_modified: Callable[[Path], None],
    ) -> None:
        """
        Initialize a watcher which observes modifications in `directory`.

        on_modified: A callable which is invoked with the path of modified
                     files within `directory`.
        """
        self.on_modified = on_modified
        self.watched_directory = str(directory)
        self.observer = Observer()

    def start(self) -> None:
        """Start watching the specified directory for file modifications."""
        event_handler = DirectoryEventHandler(self.on_modified)
        self.observer.schedule(
            event_handler,
            self.watched_directory,
            recursive=True,
        )
        try:
            self.observer.start()
        except BaseException as e:
            logger = logging.getLogger(__name__)
            logger.exception(
                'Could not start filesystem watcher.\n'
                f'Error message: "{e}".\n'
                'Set logging level to DEBUG for full stack trace.',
                exc_info=logger.getEffectiveLevel() <= logging.DEBUG,
            )

    def stop(self) -> None:
        """Stop watching the directory."""
        if self.observer.is_alive():
            try:
                self.observer.stop()
                self.observer.join()
            except (RuntimeError, SystemError):
                # TODO: Understand exactly what join() does, and why
                # it sometimes throws a RuntimeError
                # Also find out why MacOS throws SystemError
                pass


class DirectoryEventHandler(FileSystemEventHandler):
    """An event handler for filesystem changes within a directory."""

    def __init__(self, on_modified: Callable[[Path], None]) -> None:
        """Initialize event handler with callback functions."""
        self._on_modified = on_modified

    def on_modified(self, event: FileModifiedEvent) -> None:
        """Call on_modified callback function on modifed event in dir."""
        if event.is_directory:
            if platform != 'darwin':
                return

            # FSEvents on MacOS only supplies the directory containing the
            # modified file. We need to find the modified file manually...
            files_in_directory = [
                path
                for path
                in Path(event.src_path).glob('**/*')
                if not path.is_dir()
            ]
            if len(files_in_directory) > 0:
                modified_path = max(
                    files_in_directory,
                    key=lambda path: path.stat().st_mtime_ns,
                )
                self._on_modified(modified_path)
            else:
                return None
        else:
            self._on_modified(Path(event.src_path).absolute())



"""Tests for string interpolations in ModuleManager class."""

from pathlib import Path
import logging

from astrality.module import ModuleManager


def test_use_of_string_interpolations_of_module(
    tmpdir,
    caplog,
    template_directory,
):
    """Path placeholders should be replaced with compilation target."""
    temp_dir = Path(tmpdir)

    a_template = temp_dir / 'a.template'
    a_template.write_text('foobar')

    b_template = temp_dir / 'b.template'
    b_template.write_text('')
    b_target = temp_dir / 'b.target'
    b_on_modified = template_directory / 'empty.template'

    c_template = temp_dir / 'c.template'
    c_template.write_text('')
    c_target = temp_dir / 'c.target'

    modules = {
        'A': {
            'on_startup': {
                'compile': [
                    {'content': str(a_template)},
                ],
            },
        },
        'B': {
            'on_modified': {
                str(b_on_modified): {
                    'compile': [
                        {
                            'content': str(b_template),
                            'target': str(b_target),
                        },
                    ],
                },
            },
        },
        'C': {
            'on_exit': {
                'compile': {
                    'content': str(c_template),
                    'target': str(c_target),
                },
            },
        },
    }

    module_manager = ModuleManager(
        modules=modules,
        directory=temp_dir,
    )

    # Compile twice to double check that only one temporary file is inserted
    module_manager.modules['A'].execute(
        action='compile',
        block='on_startup',
    )
    module_manager.modules['A'].execute(
        action='compile',
        block='on_startup',
    )
    a_target = list(
        module_manager.modules['A'].performed_compilations().values(),
    )[0].pop()

    # Temporary compilation targets should be inserted
    assert module_manager.modules['A'].interpolate_string(
        'one two {' + str(a_template) + '}',
    ) == 'one two ' + str(a_target)

    assert module_manager.modules['A'].interpolate_string(
        '{leave/me/alone}',
    ) == '{leave/me/alone}'

    # Specified compilation targets should be inserted
    module_manager.modules['B'].execute(
        action='compile',
        block='on_modified',
        path=b_on_modified,
    )
    assert module_manager.modules['B'].interpolate_string(
        '{' + str(b_template) + '}',
    ) == str(b_target)

    # Relative paths should be resolved
    module_manager.modules['C'].execute(
        action='compile',
        block='on_exit',
    )
    assert module_manager.modules['C'].interpolate_string(
        '{c.template/into/..}',
    ) == str(c_target)

    # Check that incorrect placeholders are logged
    caplog.clear()
    module_manager.modules['C'].interpolate_string(
        '{/not/here}',
    )
    assert caplog.record_tuples == [(
        'astrality.module',
        logging.ERROR,
        'String placeholder {/not/here} could not be replaced. '
        '"/not/here" has not been compiled.',
    )]



"""Tests for TriggerAction class."""

from pathlib import Path

from astrality.actions import TriggerAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Trigger action should be a dummy when no options are provided."""
    trigger_action = TriggerAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    trigger = trigger_action.execute()
    assert trigger is None


def test_triggering_non_on_modified_block():
    """
    Triggering a startup block should return that block.

    All path attributes should be None.
    """
    trigger_action = TriggerAction(
        options={'block': 'on_startup'},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    trigger = trigger_action.execute()
    assert trigger.block == 'on_startup'
    assert trigger.specified_path is None
    assert trigger.relative_path is None
    assert trigger.absolute_path is None


def test_triggering_on_modified_block():
    """
    Triggering a on_modified block should return that block and path info.

    The path information contains the user-specified string path, the path
    relative to `directory`, and the absolute path.
    """
    trigger_action = TriggerAction(
        options={'block': 'on_modified', 'path': 'c/test.template'},
        directory=Path('/a/b'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    trigger = trigger_action.execute()
    assert trigger.block == 'on_modified'
    assert trigger.specified_path == 'c/test.template'
    assert trigger.relative_path == Path('c/test.template')
    assert trigger.absolute_path == Path('/a/b/c/test.template')



from pathlib import Path

import pytest

from astrality.module import ModuleManager


@pytest.mark.parametrize('specified_permission,expected_permission', [
    ('777', 0o777),
    ('100', 0o100),
])
def test_compiling_template_with_specific_permissions(
    test_config_directory,
    tmpdir,
    specified_permission,
    expected_permission,
):
    template = test_config_directory / 'templates' / 'empty.template'
    target = Path(tmpdir) / 'target'

    modules = {
        'test': {
            'on_startup': {
                'compile': {
                    'content': str(template),
                    'target': str(target),
                    'permissions': specified_permission,
                },
            },
        },
    }

    module_manager = ModuleManager(
        modules=modules,
    )
    module_manager.finish_tasks()

    assert (target.stat().st_mode & 0o777) == expected_permission



"""Tests for compiling YAML jinja2 templates."""

import os
from pathlib import Path
import pytest

from astrality.config import user_configuration
from astrality.utils import compile_yaml


@pytest.yield_fixture
def dir_with_compilable_files(tmpdir):
    """Create some temporary YAML files which can be compiled."""
    config_dir = Path(tmpdir)
    config_file = config_dir / 'astrality.yml'
    config_file.write_text(
        'key1: {{ env.EXAMPLE_ENV_VARIABLE }}\n'
        'key2: {{ "echo test" | shell }}',
    )

    module_file = config_dir / 'modules.yml'
    module_file.write_text(
        'key1: {{ env.EXAMPLE_ENV_VARIABLE }}\n'
        'key2: {{ "echo test" | shell }}',
    )

    yield config_dir

    os.remove(config_file)
    os.remove(module_file)
    config_dir.rmdir()


class TestUsingConfigFilesWithPlaceholders:
    def test_dict_from_config_file(self, dir_with_compilable_files):
        """Placeholders should be properly substituted."""
        config = compile_yaml(
            path=dir_with_compilable_files / 'astrality.yml',
            context={},
        )
        assert config == {
            'key1': 'test_value',
            'key2': 'test',
        }

    def test_get_user_configuration(self, dir_with_compilable_files):
        """user_configuration should use compile_yaml properly."""
        user_conf, *_ = user_configuration(dir_with_compilable_files)
        assert user_conf['key1'] == 'test_value'
        assert user_conf['key2'] == 'test'



"""Test configuration for the astrality.actions module."""

from pathlib import Path

import pytest

from astrality.config import GlobalModulesConfig


@pytest.fixture
def global_modules_config():
    return GlobalModulesConfig(
        config={},
        config_directory=Path(__file__).parent,
    )



"""Tests for astrality.xdg module."""

import os
from pathlib import Path

import pytest

from astrality.xdg import XDG


@pytest.mark.dont_patch_xdg
def test_xdg_data_home_default_location(monkeypatch):
    """Default location for XDG_DATA_HOME should be respected."""
    xdg = XDG()
    default_dir = xdg.data_home
    assert default_dir == Path('~/.local/share/astrality').expanduser()
    assert default_dir.is_dir()


@pytest.mark.dont_patch_xdg
def test_xdg_data_home_using_environment_variable(monkeypatch, tmpdir):
    """XDG_DATA_HOME environment variables should be respected."""
    custom_data_home = Path(tmpdir, 'data')
    monkeypatch.setattr(
        os,
        'environ',
        {'XDG_DATA_HOME': str(custom_data_home)},
    )

    xdg = XDG()
    data_home = xdg.data_home
    assert data_home == custom_data_home / 'astrality'
    assert data_home.is_dir()


def test_retrieving_data_resource(patch_xdg_directory_standard):
    """The data method should retrieve file resource from data home."""
    xdg = XDG()
    resource = xdg.data('modules/test.tmp')
    assert resource == patch_xdg_directory_standard / 'modules' / 'test.tmp'
    assert resource.exists()

    # We should get the same file again
    resource.write_text('hello')
    refetched_resource = xdg.data('modules/test.tmp')
    assert refetched_resource.read_text() == 'hello'


def test_retrieving_data_directory_resource(patch_xdg_directory_standard):
    """The data method should also be able to retrieve directories."""
    xdg = XDG()
    resource = xdg.data(
        resource='repositories/github',
        directory=True,
    )
    assert resource == patch_xdg_directory_standard / 'repositories' / 'github'
    assert resource.is_dir()

    # Existing directory content should not be deleted
    resource_file = resource / 'file'
    resource_file.touch()
    resource = xdg.data(
        resource='repositories/github',
        directory=True,
    )
    assert resource_file.exists()



"""Test module for all behaviour related to the import_context action."""
from astrality.context import Context


def test_importing_all_context_sections_from_file(
    test_config_directory,
    action_block_factory,
    module_factory,
    module_manager_factory,
):
    context_file = test_config_directory / 'context' / 'several_sections.yml'
    original_context = Context({
        'section2': {
            'k2_1': 'original_v2_1',
            'k2_2': 'original_v2_2',
        },
        'section3': {
            'k3_1': 'original_v3_1',
            'k3_2': 'original_v3_2',
        },
    })

    import_context = action_block_factory(
        import_context={'from_path': str(context_file)},
    )
    module = module_factory(on_startup=import_context)
    module_manager = module_manager_factory(module, context=original_context)
    assert module_manager.application_context == original_context

    # We expect the following context after import
    expected_context = Context({
        'section1': {
            'k1_1': 'v1_1',
            'k1_2': 'v1_2',
        },
        'section2': {
            'k2_1': 'v2_1',
            'k2_2': 'v2_2',
        },
        'section3': {
            'k3_1': 'original_v3_1',
            'k3_2': 'original_v3_2',
        },
    })

    # Now run startup actions, resulting in the file being imported
    module_manager.finish_tasks()

    # Assert that the new application_context is as expected
    assert module_manager.application_context == expected_context



"""Module for enforcing module 'requires' statements."""

import logging
import os
import shutil
from pathlib import Path
from typing import Union, Dict, TYPE_CHECKING, Iterable

from mypy_extensions import TypedDict

from astrality import utils


if TYPE_CHECKING:
    from astrality.module import Module  # noqa


class RequirementDict(TypedDict, total=False):
    """Available keys in requirement dictionary."""

    shell: str
    timeout: Union[int, float]
    env: str
    installed: str
    module: str


class Requirement:
    """
    Class for determining if module dependencies are satisfied.

    Object is truthy if requirements are satisfied.

    :param requirements: Dictionary containing requirements.
    :param directory: Module directory.
    :param timeout: Default timeout for shell commands.
    """

    successful: bool

    def __init__(
        self,
        requirements: RequirementDict,
        directory: Path,
        timeout: Union[int, float] = 1,
    ) -> None:
        """Construct RequirementStatement object."""
        self.successful: bool = True
        self.repr: str = ''

        # Check shell requirements
        if 'shell' in requirements:
            command = requirements['shell']
            result = utils.run_shell(
                command=command,
                fallback=False,
                timeout=requirements.get('timeout') or timeout,
                working_directory=directory,
            )
            if result is False:
                self.repr = f'Unsuccessful command: "{command}", '
                self.successful = False
            else:
                self.repr = f'Sucessful command: "{command}" (OK), '

        # Check environment requirements
        if 'env' in requirements:
            env_variable = requirements['env']
            if env_variable not in os.environ:
                self.repr += f'Missing environment variable: "{env_variable}", '
                self.successful = False
            else:
                self.repr += 'Found environment variable: ' \
                             f'"{env_variable}" (OK), '

        # Check installed requirements
        if 'installed' in requirements:
            program = requirements['installed']
            in_path = bool(shutil.which(program))
            if not in_path:
                self.repr += f'Program not installed: "{program}", '
                self.successful = False
            else:
                self.repr += f'Program installed: "{program}" (OK), '

    @staticmethod
    def pop_missing_module_dependencies(
        modules: Dict[str, 'Module'],
    ) -> Dict[str, 'Module']:
        """
        Pop modules which miss their module dependencies.

        :param modules: Dictionary containing modules.
        :return: Dictionary with removed modules which miss their module
            dependencies.
        """
        original_length = len(modules)

        for module_name, module in list(modules.items()):
            if not Requirement.satisfied_module_dependencies(
                module=module,
                enabled_modules=tuple(modules.keys()),
            ):
                del modules[module_name]
                continue

        if original_length != len(modules):
            return Requirement.pop_missing_module_dependencies(
                modules=modules,
            )

        return modules

    @staticmethod
    def satisfied_module_dependencies(
        module: 'Module',
        enabled_modules: Iterable[str],
    ) -> bool:
        """
        Return True if all module dependencies of module are enabled.

        :param module: Module to inspect for module dependencies.
        :param enabled_modules: Iterable of enabled module names.
        :return: True if all module dependencies are satisfied.
        """
        for module_dependency in module.depends_on:
            if module_dependency not in enabled_modules:
                logger = logging.getLogger(__name__)
                logger.error(
                    f'[module/{module.name}] Missing module dependency: '
                    f'"{module_dependency}". Disabling module!',
                )
                return False
        else:
            return True

    def __bool__(self) -> bool:
        """Return True if all requirements are satisfied."""
        return self.successful

    def __repr__(self) -> str:
        """Return string representation of requirement object."""
        self.repr = self.repr if self.repr else 'No requirements (OK), '
        return 'Module requirements: ' + self.repr



"""Tests for the daylight event listener subclass."""
from datetime import datetime, timedelta

import pytest

from astrality.event_listener import Daylight


@pytest.fixture
def daylight_config():
    """A daylight event listener configuration in Trondheim, Norway."""
    return {
        'type': 'daylight',
        'latitude': 0,
        'longitude': 0,
        'elevation': 0,
    }


@pytest.fixture
def daylight(daylight_config):
    """A daylight event listener in Trondheim, Norway."""
    return Daylight(daylight_config)


# --- Times around dawn ---


@pytest.fixture
def dawn(daylight):
    return daylight.construct_astral_location().sun()['dawn']


@pytest.fixture
def before_dawn(dawn):
    delta = timedelta(minutes=-2)
    return dawn + delta


@pytest.fixture
def after_dawn(dawn):
    delta = timedelta(minutes=2)
    return dawn + delta


def test_that_night_is_correctly_identified(daylight, before_dawn, freezer):
    freezer.move_to(before_dawn)
    event = daylight.event()
    assert event == 'night'


def test_that_day_is_correctly_identified(daylight, after_dawn, freezer):
    freezer.move_to(after_dawn)
    event = daylight.event()
    assert event == 'day'


# --- Times around dusk ---
@pytest.fixture
def dusk(daylight):
    return daylight.construct_astral_location().sun()['dusk']


@pytest.fixture
def before_dusk(dusk):
    delta = timedelta(minutes=-2)
    return dusk + delta


@pytest.fixture
def after_dusk(dusk):
    delta = timedelta(minutes=2)
    return dusk + delta


def test_that_night_is_correctly_identified_after_dusk(
    daylight,
    after_dusk,
    freezer,
):
    freezer.move_to(after_dusk)
    event = daylight.event()
    assert event == 'night'


def test_that_day_is_correctly_identified_before_dusk(
    daylight,
    before_dusk,
    freezer,
):
    freezer.move_to(before_dusk)
    event = daylight.event()
    assert event == 'day'


def test_time_left_before_new_event(daylight, before_dusk, freezer):
    freezer.move_to(before_dusk)
    assert daylight.time_until_next_event().total_seconds() == 120


def test_time_right_before_midnight(daylight, freezer):
    """
    This function requires special handling when the UTC time is later than all
    daylight events within the same day, which is the case right before
    midnight.
    """

    before_midnight = datetime(
        year=2019,
        month=12,
        day=23,
        hour=23,
        second=59,
        microsecond=0,
    )
    freezer.move_to(before_midnight)

    # Test that the time left is within the bounds of 0 to 24 hours
    time_left = daylight.time_until_next_event()
    assert 0 < time_left.total_seconds() < 60 * 60 * 24


def test_time_until_night_when_other_periods_are_inbetween(
    daylight,
    before_dusk,
    freezer,
):
    freezer.move_to(before_dusk - timedelta(hours=6))
    assert daylight.time_until_next_event().total_seconds() == 120 + 6 * 60 * 60



"""Tests for RunAction."""

import logging
import os
from pathlib import Path

from astrality.actions import RunAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Null objects should be executable."""
    run_action = RunAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    run_action.execute()


def test_directory_of_executed_shell_command(tmpdir):
    """All commands should be run from `directory`."""
    temp_dir = Path(tmpdir)
    run_action = RunAction(
        options={'shell': 'touch touched.tmp', 'timeout': 1},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    run_action.execute()
    assert (temp_dir / 'touched.tmp').is_file()


def test_that_dry_run_is_respected(tmpdir, caplog):
    """If dry_run is True, no commands should be executed, only logged."""
    temp_dir = Path(tmpdir)
    run_action = RunAction(
        options={'shell': 'touch touched.tmp', 'timeout': 1},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    result = run_action.execute(dry_run=True)

    # Command to be run and empty string should be returned
    assert result == ('touch touched.tmp', '')

    # Command to be run should be logged
    assert 'SKIPPED: ' in caplog.record_tuples[0][2]
    assert 'touch touched.tmp' in caplog.record_tuples[0][2]

    # Check that the command has *not* been run
    assert not (temp_dir / 'touched.tmp').is_file()


def test_use_of_replacer(tmpdir):
    """All commands should be run from `directory`."""
    temp_dir = Path(tmpdir)
    run_action = RunAction(
        options={'shell': 'whatever', 'timeout': 1},
        directory=temp_dir,
        replacer=lambda x: 'echo test',
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    command, result = run_action.execute()
    assert command == 'echo test'
    assert result == 'test'


def test_run_timeout_specified_in_action_block(tmpdir):
    """
    Run actions can time out.

    The option `timeout` overrides any timeout providided to `execute()`.
    """
    temp_dir = Path(tmpdir)
    run_action = RunAction(
        options={'shell': 'sleep 0.1 && echo hi', 'timeout': 0.05},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    _, result = run_action.execute(default_timeout=10000)
    assert result == ''

    run_action = RunAction(
        options={'shell': 'sleep 0.1 && echo hi', 'timeout': 0.2},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    _, result = run_action.execute(default_timeout=0)
    assert result == 'hi'


def test_run_timeout_specified_in_execute(tmpdir, caplog):
    """
    Run actions can time out, and should log this.

    The the option `timeout` is not specified, use `default_timeout` argument
    instead.
    """
    temp_dir = Path(tmpdir)
    run_action = RunAction(
        options={'shell': 'sleep 0.1 && echo hi'},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    _, result = run_action.execute(default_timeout=0.05)

    assert 'used more than 0.05 seconds' in caplog.record_tuples[1][2]
    assert result == ''

    run_action = RunAction(
        options={'shell': 'sleep 0.1 && echo hi'},
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    _, result = run_action.execute(default_timeout=0.2)
    assert result == 'hi'


def test_running_shell_command_with_non_zero_exit_code(caplog):
    """Shell commands with non-zero exit codes should log this."""
    run_action = RunAction(
        options={'shell': 'thiscommandshould not exist', 'timeout': 2},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    caplog.clear()
    run_action.execute()
    assert 'not found' in caplog.record_tuples[1][2]
    assert 'non-zero return code' in caplog.record_tuples[2][2]


def test_running_shell_command_with_environment_variable(caplog):
    """Shell commands should have access to the environment."""
    run_action = RunAction(
        options={'shell': 'echo $USER', 'timeout': 2},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    run_action.execute()
    assert caplog.record_tuples == [
        (
            'astrality.actions',
            logging.INFO,
            f'Running command "echo {os.environ["USER"]}".',
        ),
        (
            'astrality.utils',
            logging.INFO,
            os.environ['USER'],
        ),
    ]


def test_that_environment_variables_are_expanded():
    """String parameters in any Action type should expand env variables."""
    run_action = RunAction(
        options={'shell': 'echo $EXAMPLE_ENV_VARIABLE'},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    command, _ = run_action.execute()
    assert command == 'echo test_value'



"""Specifies everything related to application spanning configuration."""

import copy
import logging
import os
import re
from abc import ABC, abstractmethod
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import (
    Any,
    Dict,
    ClassVar,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    TYPE_CHECKING,
    Type,
    Iterable,
    Union,
)

from mypy_extensions import TypedDict

from astrality.exceptions import (
    MisconfiguredConfigurationFile,
    NonExistentEnabledModule,
)
from astrality.github import clone_repo, clone_or_pull_repo
from astrality.context import Context
from astrality import utils
from astrality.persistence import CreatedFiles
from astrality.xdg import XDG

if TYPE_CHECKING:
    from astrality.module import ModuleConfigDict  # noqa

logger = logging.getLogger(__name__)


class EnablingStatementRequired(TypedDict):
    """Required items of astrality.yml::modules:enabled_modules."""

    name: str


class EnablingStatement(EnablingStatementRequired, total=False):
    """Optional items of astrality.yml::modules:enabled_modules."""

    trusted: bool
    autoupdate: bool


class GlobalModulesConfigDict(TypedDict, total=False):
    """Optional items in astrality.yml::modules."""

    requires_timeout: Union[int, float]
    run_timeout: Union[int, float]
    reprocess_modified_files: bool
    modules_directory: str
    enabled_modules: List[EnablingStatement]


class GlobalAstralityConfigDict(TypedDict, total=False):
    """Optional items in astrality.yml::astrality."""

    hot_reload_config: bool
    startup_delay: Union[int, float]


class AstralityYAMLConfigDict(TypedDict, total=False):
    """Optional items in astrality.yml."""

    astrality: GlobalAstralityConfigDict
    modules: GlobalModulesConfigDict


ASTRALITY_DEFAULT_GLOBAL_SETTINGS: AstralityYAMLConfigDict = {
    'astrality': {
        'hot_reload_config': False,
        'startup_delay': 0,
    },
    'modules': {
        'requires_timeout': 1,
        'run_timeout': 0,
        'reprocess_modified_files': False,
        'modules_directory': 'modules',
        'enabled_modules': [
            {'name': '*'},
            {'name': '*::*'},
        ],
    },
}


def resolve_config_directory() -> Path:
    """
    Return the absolute configuration directory path for the application.

    The directory path is resolved as follows:
        1) If $ASTRALITY_CONFIG_HOME is present, use it.
        2) If $XDG_CONFIG_HOME is present, use $XDG_CONFIG_HOME/astrality.
        3) Elsewise, use ~/config/astrality.
    """
    if 'ASTRALITY_CONFIG_HOME' in os.environ:
        # The user has set a custom config directory for astrality
        config_directory = Path(os.environ['ASTRALITY_CONFIG_HOME'])
    else:
        # Follow the XDG directory standard
        config_directory = Path(
            os.getenv('XDG_CONFIG_HOME', '~/.config'),
            'astrality',
        )
    return config_directory.expanduser().absolute()


def infer_config_location(
    config_directory: Optional[Path] = None,
) -> Tuple[Path, Path]:
    """
    Return path of Astrality configuration file based on config folder path.

    Try to find the configuration directory and file for astrality, based on
    filesystem or specific environment variables if they are present several
    places to put it. If the expected config file is not present, use an
    example configuration instead.
    """
    config_directory = config_directory or resolve_config_directory()
    config_file = config_directory / 'astrality.yml'

    if not config_file.is_file():
        logger.warning(
            'Configuration file not found in its expected path "{config_file}.'
            'Using default values for astrality.yml.',
        )
    else:
        logging.info(f'Using configuration file "{config_file}"')

    return config_directory, config_file


def user_configuration(
    config_directory: Optional[Path] = None,
) -> Tuple[
    AstralityYAMLConfigDict,
    Dict[str, 'ModuleConfigDict'],
    Context,
    Path,
]:
    """
    Return instantiation parameters for ModuleManager.

    :return: Tuple containing: astrality.yml dictionary, global modules
        dictionary, global context dictionary, and path to config directory.
    """
    config_directory, config_file = infer_config_location(config_directory)

    # First get global context, which we can use when compiling other files
    context_file = config_directory / 'context.yml'
    if context_file.exists():
        global_context = Context(utils.compile_yaml(
            path=context_file,
            context=Context(),
        ))
    else:
        global_context = Context()

    # Global configuration options
    config: AstralityYAMLConfigDict
    if config_file.is_file():
        config = utils.compile_yaml(  # type: ignore
            path=config_file,
            context=global_context,
        )
    else:
        config = {}

    # Insert default global settings that are not specified
    for section_name in ('astrality', 'modules'):
        section_content = config.get(section_name, {})
        config[section_name] = ASTRALITY_DEFAULT_GLOBAL_SETTINGS[section_name].copy()  # type: ignore # noqa
        config[section_name].update(section_content)  # type: ignore

    # Globally defined modules
    modules_file = config_directory / 'modules.yml'
    if modules_file.exists():
        modules = utils.compile_yaml(
            path=modules_file,
            context=global_context,
        )
    else:
        modules = {}

    return config, modules, global_context, config_directory


def create_config_directory(path: Optional[Path] = None, empty=False) -> Path:
    """Create application configuration directory and return its path."""
    if not path:
        path = resolve_config_directory()

    if not path.exists():
        if empty:
            logger.warning(f'Creating empty directory at "{str(path)}".')
            path.mkdir(parents=True)
        else:
            logger.warning(
                f'Copying over example config directory to "{str(path)}".',
            )
            example_config_dir = Path(__file__).parent / 'config'
            copy_tree(
                src=str(example_config_dir),
                dst=str(path),
            )
    else:  # pragma: no cover
        logger.warning(f'Path "{str(path)}" already exists! Delete it first.')

    return path


def expand_path(path: Path, config_directory: Path) -> Path:
    """
    Return an absolute path from a (possibly) relative path.

    Relative paths are relative to $ASTRALITY_CONFIG_HOME, and ~ is
    expanded to the home directory of $USER.
    """
    # Expand environment variables present in path
    path = Path(os.path.expandvars(path))

    # Expand any tilde expressions for user home directory
    path = path.expanduser()

    # Use config directory as anchor for relative paths
    if not path.is_absolute():
        path = Path(os.path.expandvars(config_directory)) / path

    # Return path where symlinks such as '..' are resolved
    return path.resolve()


def expand_globbed_path(path: Path, config_directory: Path) -> Set[Path]:
    """
    Expand globs, i.e. * and **, of path object.

    This function is actually not used at the moment, but I have left it here
    in case we would want to support globbed paths in the future.

    :param path: Path to be expanded.
    :param config_directory: Anchor for relative paths.
    :return: Set of file paths resulting from glob expansion.
    """
    # Make relative paths absolute with config_directory as anchor
    path = expand_path(path=path, config_directory=config_directory)

    # Remove root directory from path
    relative_to_root = Path(*path.parts[1:])

    # Expand all globs in the entirety of `path`, recursing if ** is present
    expanded_paths = Path('/').glob(
        pattern=str(relative_to_root),
    )

    # Cast generator to set, and remove directories
    return set(path for path in expanded_paths if path.is_file())


class ModuleSource(ABC):
    """Superclass for the source of an enabled module."""

    # Directory containing "modules.yml" and "context.yml"
    directory: Path

    # Path to "modules.yml"
    modules_file: Path

    # Path to "context.yml"
    context_file: Path

    # Name syntax defining how to enable this module source type
    name_syntax: ClassVar[Pattern]

    # What to prepend to module names in order to avoid name collisions
    prepend: str

    # String specifying which modules that are enabled in module source
    enabled_module: str

    # Cached property containing module configurations
    _modules: Dict[str, 'ModuleConfigDict']

    # Cached property containing module context
    _context: Context

    # Options defined in astrality.yml
    _config: Dict[str, 'ModuleConfigDict']

    @abstractmethod
    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """Initialize a module source from an enabling statement."""
        raise NotImplementedError

    def modules(self, context: Context) -> Dict[Any, Any]:
        """
        Return modules defined in modules source.

        :param context: Context used when compiling "modules.yml".
        :return: Modules dictionary, with module name keys prepended with
            '/module', and module configuration values.
        """
        if not self.modules_file.exists():
            self._modules = {}

        if hasattr(self, '_modules'):
            return self._modules

        self._modules = filter_config_file(
            config_file=self.modules_file,
            context=context,
            enabled_module_name=self.enabled_module,
            prepend=self.prepend,
        )
        return self._modules

    def context(self, context: Context = Context()) -> Context:
        """
        Return context defined in module source.

        :param context: Context used when compiling "context.yml".
        :return: Context dictionary.
        """
        if not self.context_file.exists():
            return Context()

        if hasattr(self, '_context'):
            return self._context

        self._context = Context(utils.compile_yaml(
            path=self.context_file,
            context=context,
        ))
        return self._context

    def config(
        self,
        context: Context = Context(),
    ) -> Dict[str, 'ModuleConfigDict']:
        """
        Return all configuration options defined in module source.

        This includes modules (prepended with 'module/') and context (prepended
        with 'context/'.

        :param context: Context used when compiling "modules.yml" and
            "context.yml".
        :return: Module and context dictionary.
        """
        if hasattr(self, '_config'):
            return self._config

        self._config = self.modules(context=context)
        return self._config

    def __contains__(self, module_name: str) -> bool:
        """Return True if this source contains enabled module_name."""
        return module_name in self._modules

    @classmethod
    def represented_by(cls, module_name: str) -> bool:
        """Return True if name represents module source type."""
        return bool(cls.name_syntax.match(module_name))

    @classmethod
    def type(cls, of: str) -> 'Type[ModuleSource]':
        """Return the subclass which is responsible for the module name."""
        for source_type in cls.__subclasses__():
            if source_type.represented_by(module_name=of):
                return source_type

        raise MisconfiguredConfigurationFile(  # pragma: no cover
            f'Tried to enable invalid module name syntax "{of}".',
        )


class GlobalModuleSource(ModuleSource):
    """Module defined in `$ASTRALITY_CONFIG_HOME/modules.yml`."""

    name_syntax = re.compile(r'^(\w+|\*)$')
    prepend = ''

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """Initialize enabled module defined in `astrality.yml`."""
        self.enabled_module = enabling_statement['name']
        assert self.represented_by(module_name=self.enabled_module)

        assert modules_directory.is_absolute()
        self.directory = modules_directory
        self.modules_file = modules_directory / 'modules.yml'
        self.context_file = modules_directory / 'context.yml'

    def __contains__(self, module_name: str) -> bool:
        """Return True if the module name is enabled."""
        if self.enabled_module == '*' and self.represented_by(module_name):
            return True
        else:
            return module_name == self.enabled_module

    def __repr__(self) -> str:
        """Return the string representation of the global module source."""
        return f"GlobalModuleSource('{self.enabled_module}')"


class GithubModuleSource(ModuleSource):
    """Module defined in a GitHub repository."""

    name_syntax = re.compile(r'^github::.+/.+(::(\w+|\*))?$')

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """
        Initialize enabled module defined in Github repository.

        A GitHub module is enabled with the name syntax:
        github::<github_user>/<github_repo>[::<enabled_module>]

        If <enabled_module> is skipped, all modules are enabled, i.e.
        github::<github_user>/<github_repo>::*
        """
        assert self.represented_by(
            module_name=enabling_statement['name'],
        )
        assert modules_directory.is_absolute()

        specified_module = enabling_statement['name']
        self.autoupdate = enabling_statement.get('autoupdate', False)

        github_path, *enabled_modules = specified_module[8:].split('::')
        if len(enabled_modules) > 0:
            self.enabled_module = enabled_modules[0]
        else:
            self.enabled_module = '*'

        self.github_user, self.github_repo = github_path.split('/')
        self.prepend = f'github::{self.github_user}/{self.github_repo}::'

        repositories = XDG().data(
            resource='repositories/github',
            directory=True,
        )
        self.directory = repositories / github_path
        self.modules_file = self.directory / 'modules.yml'
        self.context_file = self.directory / 'context.yml'

        if not self.directory.is_dir():
            clone_repo(
                user=self.github_user,
                repository=self.github_repo,
                modules_directory=repositories,
            )
        elif self.autoupdate:
            clone_or_pull_repo(
                user=self.github_user,
                repository=self.github_repo,
                modules_directory=repositories,
            )

    def __eq__(self, other) -> bool:
        """
        Return True if being compared to identical github module.

        Return True if two GithubModuleSource objects represent the same
        directory with identical modules enabled.
        """
        try:
            return self.directory == other.directory \
                and self._config == other._config
        except AttributeError:
            return False

    def __repr__(self) -> str:
        """Return string representation of the github module source."""
        return f"GithubModuleSource('{self.github_user}/{self.github_repo}')" \
            f' = {tuple(self._config.keys())}'


class DirectoryModuleSource(ModuleSource):
    """
    Module(s) defined in a directory.

    Specifically: `$ASTRALITY_CONFIG_HOME/{modules_directory}/config.yml
    """

    name_syntax = re.compile(r'^(?!github::).+::(\w+|\*)$')

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """Initialize an DirectoryModuleSource object."""
        assert self.represented_by(
            module_name=enabling_statement['name'],
        )
        assert modules_directory.is_absolute()

        relative_directory_path, self.enabled_module = \
            enabling_statement['name'].split('::')
        self.relative_directory_path = Path(relative_directory_path)
        self.prepend = str(self.relative_directory_path) + '::'
        self.trusted = enabling_statement.get('trusted', True)

        self.directory = expand_path(
            path=self.relative_directory_path,
            config_directory=modules_directory,
        )
        self.modules_file = self.directory / 'modules.yml'
        self.context_file = self.directory / 'context.yml'

    def __repr__(self):
        """Human-readable representation of a DirectoryModuleSource object."""
        return ''.join((
            'DirectoryModuleSource(',
            f'name={self.relative_directory_path}::',
            f'{self.enabled_module_name}, ',
            f'directory={self.directory}, ',
            f'trusted={self.trusted})',
        ))

    def __eq__(self, other) -> bool:
        """
        Return true if compared to identical module defined in a directory.

        Entirily determined by the source directory of the module.
        """
        try:
            return self.directory == other.directory
        except AttributeError:  # pragma: no cover
            return False


class EnabledModules:
    """
    Object which keeps track of all enabled modules.

    Also allows fetching of configurations defined externally.
    """

    def __init__(
        self,
        enabling_statements: List[EnablingStatement],
        config_directory: Path,
        modules_directory: Path,
    ) -> None:
        """Determine exactly which modules are enabled from user config."""
        enabling_statements = self.process_enabling_statements(
            enabling_statements=enabling_statements,
            modules_directory=modules_directory,
        )

        self.source_types: Dict[Type[ModuleSource], List[ModuleSource]] = {
            GlobalModuleSource: [],
            DirectoryModuleSource: [],
            GithubModuleSource: [],
        }

        for enabling_statement in enabling_statements:
            try:
                source_type = ModuleSource.type(of=enabling_statement['name'])
            except MisconfiguredConfigurationFile:
                logger.error(
                    f'Invalid module name syntax {str(enabling_statement)} '
                    'in enabled_modules configuration.',
                )
                continue

            if source_type == GlobalModuleSource:
                source_directory = config_directory
            else:
                source_directory = modules_directory

            self.source_types[source_type].append(source_type(
                enabling_statement=enabling_statement,
                modules_directory=source_directory,
            ))

    def process_enabling_statements(
        self,
        enabling_statements: List[EnablingStatement],
        modules_directory: Path,
    ):
        """Return enabling statements where wildcards have been replaced."""
        self.all_global_modules_enabled = False
        self.all_directory_modules_enabled = False

        new_enabling_statements: List[EnablingStatement] = []
        for enabling_statement in enabling_statements:
            if enabling_statement['name'] == '*':
                self.all_global_modules_enabled = True
                new_enabling_statements.append(enabling_statement)

            elif enabling_statement['name'] == '*::*':
                self.all_directory_modules_enabled = True

                for module_directory in self.module_directories(
                    within=modules_directory,
                ):
                    directory_module = module_directory + '::*'
                    new_enabling_statement = copy.deepcopy(enabling_statement)
                    new_enabling_statement['name'] = directory_module
                    new_enabling_statements.append(new_enabling_statement)

            else:
                new_enabling_statements.append(enabling_statement)

        return new_enabling_statements

    @staticmethod
    def module_directories(within: Path) -> Tuple[str, ...]:
        """Return all subdirectories which contain module definitions."""
        try:
            return tuple(
                path.name
                for path
                in within.glob('**/*')
                if (path / 'modules.yml').is_file() or
                (path / 'context.yml').is_file()
            )
        except FileNotFoundError:
            logger.error(
                f'Tried to search for module directories in "{within}", '
                'but directory does not exist!.',
            )
            return ()

    def compile_config_files(
        self,
        context: Context,
    ):
        """Compile all config templates with context."""
        for source in (
            *self.source_types[DirectoryModuleSource],
            *self.source_types[GithubModuleSource],
        ):
            source.config(context=context)

    def __contains__(self, module_name: str) -> bool:
        """Return True if the given module name is supposed to be enabled."""
        source_type = ModuleSource.type(of=module_name)
        for module_source in self.source_types[source_type]:
            if module_name in module_source:
                return True

        return False

    def __repr__(self) -> str:
        """Return string representation of all enabled modules."""
        return ', '.join(
            source.__repr__()
            for source
            in self.source_types.values()
        )


class GlobalModulesConfig:
    """
    User modules configuration.

    The plan is to make this the input argument for ModuleManager.__init__(),
    extracting all logic related to getting module configurations (only enebled
    ones) from the object.
    """

    def __init__(
        self,
        config: GlobalModulesConfigDict,
        config_directory: Path,
    ) -> None:
        """Initialize a GlobalModulesConfig object from a dictionary."""
        self.reprocess_modified_files = config.get(
            'reprocess_modified_files',
            False,
        )
        self.requires_timeout = config.get(
            'requires_timeout',
            1,
        )
        self.run_timeout = config.get(
            'run_timeout',
            0,
        )
        self.created_files = CreatedFiles()

        # Determine the directory which contains external modules
        assert config_directory.is_absolute()
        self.config_directory = config_directory
        if 'modules_directory' in config:
            # Custom modules folder
            modules_path = expand_path(
                path=Path(config['modules_directory']),
                config_directory=self.config_directory,
            )
            self.modules_directory = modules_path
        else:
            # Default modules folder: $ASTRALITY_CONFIG_HOME/modules
            self.modules_directory = config_directory / 'modules'

        # Enable all modules if nothing is specified
        self.enabled_modules = EnabledModules(
            enabling_statements=config.get(
                'enabled_modules',
                [
                    {'name': '*', 'trusted': True},
                    {'name': '*::*', 'trusted': True},
                ],
            ),
            config_directory=self.config_directory,
            modules_directory=self.modules_directory,
        )

    @property
    def external_module_sources(
        self,
    ) -> Iterable[ModuleSource]:
        """Return an iterator yielding all external module configs."""
        for source in (
            *self.enabled_modules.source_types[DirectoryModuleSource],
            *self.enabled_modules.source_types[GithubModuleSource],
        ):
            yield source

    def compile_config_files(
        self,
        context: Context,
    ):
        """Compile all config templates with context."""
        self.enabled_modules.compile_config_files(context)


def filter_config_file(
    config_file: Path,
    context: Context,
    enabled_module_name: str,
    prepend: str,
) -> Dict[str, 'ModuleConfigDict']:
    """
    Return a filtered dictionary representing `config_file`.

    Only modules given by `enabled_module_name` are kept, and their names
    are prepended with `prepend`.
    """
    assert config_file.name == 'modules.yml'

    try:
        modules_dict = utils.compile_yaml(
            path=config_file,
            context=context,
        )
    except FileNotFoundError:
        logger.warning(
            f'Non-existent module configuration file "{config_file}" '
            'Skipping enabled module '
            f'"{prepend}{enabled_module_name}"',
        )
        return {}

    if not modules_dict:  # pragma: no cover
        logger.warning(
            f'Empty modules configuration "{config_file}".',
        )
        return {}
    elif not isinstance(modules_dict, dict):  # pragma: no cover
        logger.critical(
            f'Configuration file "{config_file}" not formated as '
            'a dictionary at root indentation.',
        )
        raise MisconfiguredConfigurationFile

    modules = tuple(modules_dict.keys())
    if enabled_module_name != '*' \
            and enabled_module_name not in modules:
        raise NonExistentEnabledModule

    # We rename each module to module/{self.name}.module_name
    # in order to prevent naming conflicts when using modules provided
    # from a third party with the same name as another managed module.
    # This way you can use a module named "conky" from two third parties,
    # in addition to providing your own.
    for module_name in modules:
        if not enabled_module_name == '*' \
           and enabled_module_name != module_name:
            # The module is not enabled, remove the module
            modules_dict.pop(module_name)
            continue

        # Replace the module name with folder_name.module_name
        non_conflicting_module_name = prepend + module_name
        module_section = modules_dict.pop(module_name)
        modules_dict[non_conflicting_module_name] = module_section

    return modules_dict



"""Module which keeps track of module setup block actions and created files."""

import hashlib
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable, List, Optional

from mypy_extensions import TypedDict

from astrality import actions, utils
from astrality.xdg import XDG


class CreationMethod(Enum):
    """Ways modules can create files."""
    COMPILE = 'compiled'
    COPY = 'copied'
    MKDIR = 'mkdir'
    SYMLINK = 'symlinked'


class CreationInfo(TypedDict):
    """Information stored for each module created file."""
    # Source file used to create file
    content: str

    # Method used to create file
    method: str

    # Last modification MD5 hash of created file
    # Set to None if PermissionError
    hash: Optional[str]

    # Possible backup of replaced existing file
    backup: Optional[str]


# Contents of $XDG_DATA_HOME/astrality/created_files.yml.
# Example: {'module_name': {'/created/file': {...}, '/another/file': {...}}}
CreationsYAML = Dict[str, Dict[str, CreationInfo]]


class CreatedFiles:
    """Object which persists which files that have been created by modules."""

    # Dictionary read from created_files.yml containing files created by modules
    creations: CreationsYAML

    # Path to file containing module created files
    _path: Path

    def __init__(self) -> None:
        """Constuct CreatedFiles object."""
        self.creations = utils.load_yaml(path=self.path)

    def wrapper_for(self, module: str) -> 'ModuleCreatedFiles':
        """
        Return CreatedFiles object wrapper for specific module.

        This allows you to not pass the module name when calling CreatedFiles
        methods. This method is used by the ActionBlock object when creating
        Action instances, allowing Action instances to save created files
        without knowing which module it is part of.

        :param module: Name of module that will create files with this object.
        """
        return ModuleCreatedFiles(creation_store=self, module=module)

    @property
    def path(self) -> Path:
        """Return path to file which stores files created by modules."""
        if hasattr(self, '_path'):
            return self._path

        xdg = XDG('astrality')
        self._path = xdg.data(resource='created_files.yml')
        if os.stat(self._path).st_size == 0:
            self._path.touch()
            utils.dump_yaml(data={}, path=self._path)

        return self._path

    def insert(
        self,
        module: str,
        creation_method: CreationMethod,
        contents: Iterable[Optional[Path]],
        targets: Iterable[Path],
    ) -> None:
        """
        Insert files created by a module.

        :param module: Name of module which has created the files.
        :param creation_method: Type of action which has created the file.
        :param contents: The source files used in creating the files. None if
          a created directory.
        :param targets: The files that have be created.
        """
        # We do not want to insert empty sections, to reduce reduntant clutter
        if not contents:
            return

        modified = False
        module_section = self.creations.setdefault(module, {})

        for content, target in zip(contents, targets):
            # Do not insert files that actually do not exist
            if not target.exists():
                continue

            creation = module_section.setdefault(
                str(target),
                {},  # type: ignore
            )
            if creation.get('content') != str(content):
                modified = True
                creation['content'] = str(content)
                creation['method'] = creation_method.value
                creation.setdefault('backup', None)  # type: ignore

                try:
                    creation['hash'] = hashlib.md5(
                        target.read_bytes(),
                    ).hexdigest()
                except (PermissionError, IsADirectoryError):
                    # Currently, we do not hash directories or content with
                    # lacking permissions to read.
                    creation['hash'] = None

        if modified:
            utils.dump_yaml(data=self.creations, path=self.path)

    def by(self, module) -> List[Path]:
        """
        Return files created by module.

        :param module: Name of module.
        :return: List of paths to created files.
        """
        return [
            Path(creation)
            for creation
            in self.creations.get(module, {}).keys()
        ]

    def cleanup(self, module: str, dry_run: bool = False) -> None:
        """
        Delete files created by module.

        :param module: Name of module, file creation of which will be deleted.
        :param dry_run: If True, no files will be deleted, only logging will
            occur.
        """
        logger = logging.getLogger(__name__)
        module_creations = self.creations.get(module, {})

        # This dictionary will be populated with all directories which we can't
        # delete. Those should still be tracked after cleaning up the module.
        dangling_directories: Dict[str, CreationInfo] = {}

        for creation, info in sorted(
            module_creations.items(),
            key=lambda item: -len(Path(item[0]).parts),  # depth-first order
        ):
            creation_method = info['method']
            content = info['content']
            backup = info['backup']
            log_msg = (
                f'[Cleanup] Deleting "{creation}" '
                f'({creation_method} content from "{content}"). '
                f'Backup replacement: {backup}.'
            )
            if dry_run:
                logger.info('SKIPPED: ' + log_msg)
                continue

            creation_path = Path(creation)
            if creation_path.is_dir():
                try:
                    logger.info(log_msg)
                    creation_path.rmdir()
                except OSError:
                    logger.warning(
                        f'Failed to remove created directory "{creation}", '
                        'as it contains new non-module files since creation! '
                        'Try to delete files manually and then cleanup again.',
                    )
                    dangling_directories[creation] = info
            elif creation_path.exists():
                logger.info(log_msg)
                creation_path.unlink()
            else:
                logger.info(log_msg + ' [No longer exists!]')

            if backup and Path(backup).exists():
                utils.move(  # type: ignore
                    source=info['backup'],
                    destination=creation,
                    follow_symlinks=False,
                )

        if not dry_run:
            self.creations.pop(module, None)
            if dangling_directories:
                self.creations[module] = dangling_directories
            utils.dump_yaml(data=self.creations, path=self.path)

    def backup(self, module: str, path: Path) -> Optional[Path]:
        """
        Take backup of path if it exists and is not created by Astrality.

        :param module: Module requesting file to be backed up.
        :param path: Path to file to back up.
        :return: Optional path to backup file.
        """
        if path in self or not path.exists():
            return None

        filepath_hash = hashlib.md5(
            str(path).encode('utf-8'),
        ).hexdigest()[:7]
        backup_filename = path.name + '-' + filepath_hash
        backup = XDG().data(f'backups/{module}/{backup_filename}')
        utils.move(
            source=path,
            destination=backup,
            follow_symlinks=False,
        )

        self.creations.setdefault(module, {})[str(path)] = {  # type: ignore
            'backup': str(backup),
        }
        utils.dump_yaml(data=self.creations, path=self.path)
        return backup

    def mkdir(self, module: str, path: Path) -> None:
        """
        Create directory with parents and persist the creation.

        A no-op if the directory already exists. All parent directories that
        also needs to be created will also be persisted.

        :param module: Module that requests the directory to be created.
        :param path: Directory path to be made.
        """
        # First find out exactly which directories, including parents, will
        # be created.
        created_directories = []
        for new_directory in (path, *path.parents):
            if not new_directory.exists():
                created_directories.append(new_directory)
            else:
                break

        path.mkdir(parents=True, exist_ok=True)
        self.insert(
            module=module,
            creation_method=CreationMethod.MKDIR,
            contents=len(created_directories) * [None],
            targets=created_directories,
        )

    def __contains__(self, path) -> bool:
        """Return True if path has been created by Astrality."""
        return any(
            str(path) in module_creations.keys()
            for module_creations
            in self.creations.values()
        )

    def __repr__(self) -> str:
        """Return string representation of CreatedFiles object."""
        return f'CreatedFiles(path={self.path})'


class ModuleCreatedFiles:
    """Wrapper Class for managing created files by specific module."""

    def __init__(self, creation_store: CreatedFiles, module: str) -> None:
        """
        Construct CreatedFiles object wrapper for specific module.

        :param creation_store: Created file object instance.
        :param module: Name of module using object to persist created files.
        """
        self.module = module
        self.creation_store = creation_store

    def backup(self, path: Path) -> Optional[Path]:
        """
        Backup path if not created by Astrality.

        :param path: Path to file to be backed up.
        :return: Optional path to backup.
        """
        return self.creation_store.backup(module=self.module, path=path)

    def mkdir(self, path: Path) -> None:
        """
        Create directory with parents and persist the creation.

        A no-op if the directory already exists.

        :param path: Directory path to be made.
        """
        self.creation_store.mkdir(module=self.module, path=path)

    def insert_creation(
        self,
        content: Path,
        target: Path,
        method: CreationMethod,
    ) -> None:
        """
        Persist file created by self.module.

        :param content: Path to content used to create new file.
        :param target: Path to created file.
        :param method: Action method used to create file.
        """
        self.creation_store.insert(
            module=self.module,
            contents=[content],
            targets=[target],
            creation_method=method,
        )


class ExecutedActions:
    """
    Object which persists executed module actions.

    :param module_name: Unique string id of module.
    """

    # Path to file containing executed setup actions
    _path: Path

    def __init__(self, module_name: str) -> None:
        """Construct ExecutedActions object."""
        self.module = module_name
        self.new_actions = {}  # type: ignore

        file_data = utils.load_yaml(path=self.path)
        self.old_actions = file_data.setdefault(
            self.module,
            {},
        )

    def is_new(
        self,
        action_type: str,
        action_options,
    ) -> bool:
        """
        Return True if the specified action has not been executed earlier.

        :param action_type: Type of action, see ActionBlock.action_types.
        :param action_options: Configuration of action to be performed.
        """
        assert action_type in actions.ActionBlock.action_types
        if not action_options:
            # Empty actions can be disregarded.
            return False

        is_new = action_options not in self.old_actions.get(
            action_type,
            [],
        ) and action_options not in self.new_actions.get(
            action_type,
            [],
        )

        if is_new:
            self.new_actions                 \
                .setdefault(action_type, []) \
                .append(action_options)

        return is_new

    def write(self) -> None:
        """Persist all actions that have been checked in object lifetime."""
        if not self.new_actions:
            return

        file_data = utils.load_yaml(path=self.path)
        file_data.setdefault(self.module, {})

        for action_type, action_options in self.new_actions.items():
            file_data[self.module].setdefault(
                action_type,
                [],
            ).extend(action_options)

        utils.dump_yaml(
            path=self.path,
            data=file_data,
        )

    def reset(self) -> None:
        """Delete all executed module actions."""
        file_data = utils.load_yaml(path=self.path)
        reset_actions = file_data.pop(self.module, None)

        logger = logging.getLogger(__name__)
        if not reset_actions:
            logger.error(
                'No saved executed on_setup actions for module '
                f'"{self.module}"!',
            )
        else:
            logger.info(
                f'Reset the following actions for module "{self.module}":\n' +
                utils.yaml_str({self.module: reset_actions}),
            )

        utils.dump_yaml(
            path=self.path,
            data=file_data,
        )
        self.old_actions = {}

    @property
    def path(self) -> Path:
        """Return path to file which stores executed module setup actions."""
        if hasattr(self, '_path'):
            return self._path

        xdg = XDG('astrality')
        self._path = xdg.data(resource='setup.yml')
        if os.stat(self._path).st_size == 0:
            self._path.touch()
            utils.dump_yaml(data={}, path=self._path)

        return self._path

    def __repr__(self) -> str:
        """Return string representation of ExecutedActions object."""
        return f'ExecutedActions(module_name={self.module}, path={self.path})'



"""Tests for module requirements."""

import logging
from pathlib import Path

from astrality.module import Module, ModuleManager
from astrality.tests.utils import RegexCompare


def test_module_requires_option(caplog):
    """Test that modules are disabled when they don't satisfy `requires`."""
    # Simple module that satisfies the requirements
    does_satisfy_requiremnets = {
        'enabled': True,
        'requires': {'shell': 'command -v cd'},
    }
    assert Module.valid_module(
        name='satisfies',
        config=does_satisfy_requiremnets,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )

    # Simple module that does not satisfy requirements
    does_not_satisfy_requirements = {
        'requires': {'shell': 'command -v does_not_exist'},
    }
    assert not Module.valid_module(
        name='does_not_satisfy',
        config=does_not_satisfy_requirements,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )
    assert (
        'astrality.module',
        logging.WARNING,
        '[module/does_not_satisfy] Module requirements: '
        'Unsuccessful command: "command -v does_not_exist", !',
    ) in caplog.record_tuples

    # Test failing one of many requirements
    does_not_satisfy_one_requirement = {
        'requires': [
            {'shell': 'command -v cd'},
            {'shell': 'command -v does_not_exist'},
        ],
    }
    caplog.clear()
    assert not Module.valid_module(
        name='does_not_satisfy',
        config=does_not_satisfy_one_requirement,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )
    assert (
        'astrality.module',
        logging.WARNING,
        RegexCompare(
            r'\[module/does_not_satisfy\] Module requirements: .+ '
            'Unsuccessful command: "command -v does_not_exist", !',
        ),
    ) in caplog.record_tuples


def test_module_module_dependencies():
    """ModuleManager should remove modules with missing module dependencies."""
    config = {
        'modules': {
            'modules_directory': 'freezed_modules',
            'enabled_modules': [
                {'name': 'north_america::*'},
                {'name': 'A'},
                {'name': 'B'},
                {'name': 'C'},
            ],
        },
    }
    modules = {
        'A': {'requires': {'module': 'north_america::USA'}},
        'B': {'requires': [{'module': 'A'}]},
        'C': {'requires': [{'module': 'D'}]},
    }

    module_manager = ModuleManager(
        config=config,
        modules=modules,
    )
    assert sorted(module_manager.modules.keys()) \
        == sorted(['A', 'B', 'north_america::USA'])



#!/usr/bin/env python3.6

"""The module meant to be run in order to start Astrality."""

import logging
import os
import signal
import sys
import time
from typing import List

import psutil

from astrality import utils
from astrality.config import user_configuration
from astrality.module import ModuleManager
from astrality.xdg import XDG


logger = logging.getLogger(__name__)


def main(
    modules: List[str] = [],
    logging_level: str = 'INFO',
    dry_run: bool = False,
    test: bool = False,
):
    """
    Run the main process for Astrality.

    :param modules: Modules to be enabled. If empty, use astrality.yml.
    :param logging_level: Loging level.
    :param dry_run: If file system actions should be printed and skipped.
    :param test: If True, return after one iteration loop.
    """
    if 'ASTRALITY_LOGGING_LEVEL' in os.environ:
        # Override logging level if env variable is set
        logging_level = os.environ['ASTRALITY_LOGGING_LEVEL']

    # Set the logging level to the configured setting
    logging.basicConfig(
        level=logging.getLevelName(logging_level),  # type: ignore
    )

    if not modules and not dry_run and not test:
        # Quit old astrality instances
        kill_old_astrality_processes()

    # How to quit this process
    def exit_handler(signal=None, frame=None) -> None:
        """
        Cleanup all temporary files and run module exit handlers.

        The temp directory is left alone, for two reasons:
        1: An empty directory uses neglible disk space.
        2: If this process is interrupted by another Astrality instance,
           we might experience race conditions when the exit handler deletes
           the temporary directory *after* the new Astrality instance creates
           it.
        """
        logger.critical('Astrality was interrupted')
        logger.info('Cleaning up temporary files before exiting...')

        try:
            # Run all the module exit handlers
            module_manager.exit()
        except NameError:
            # The module_manager instance has not been assigned yet.
            pass

        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)

    # Some SIGINT signals are not properly interupted by python and converted
    # into KeyboardInterrupts, so we have to register a signal handler to
    # safeguard against such cases. This seems to be the case when conky is
    # launched as a subprocess, making it the process that receives the SIGINT
    # signal and not python. These signal handlers cause issues for \
    # NamedTemporaryFile.close() though, so they are only registrered when
    # we are not testing.
    if not test:
        signal.signal(signal.SIGINT, exit_handler)

        # Also catch kill-signkal from OS,
        # e.g. `kill $(pgrep -f "python astrality.py")`
        signal.signal(signal.SIGTERM, exit_handler)

    try:
        (
            config,
            module_configs,
            global_context,
            directory,
        ) = user_configuration()

        if modules:
            config['modules']['enabled_modules'] = [
                {'name': module_name}
                for module_name
                in modules
            ]

        # Delay further actions if configuration says so
        time.sleep(config['astrality']['startup_delay'])

        module_manager = ModuleManager(
            config=config,
            modules=module_configs,
            context=global_context,
            directory=directory,
            dry_run=dry_run,
        )
        module_manager.finish_tasks()

        while True:
            if module_manager.has_unfinished_tasks():
                # TODO: Log which new event which has been detected
                logger.info('New event detected.')
                module_manager.finish_tasks()
                logger.info(f'Event change routine finished.')

            if test or dry_run:
                logger.debug('Main loop interupted due to --dry-run.')
                return
            elif not module_manager.keep_running:
                logger.info(
                    'No more tasks to be performed. '
                    'Executing on_exit blocks.',
                )
                module_manager.exit()
                return
            else:
                logger.info(
                    f'Waiting {module_manager.time_until_next_event()} '
                    'until next event change and ensuing update.',
                )

                # Weird bug related to sleeping more than 10e7 seconds
                # on MacOS, causing OSError: Invalid Argument
                wait = module_manager.time_until_next_event().total_seconds()
                if wait >= 10e7:
                    wait = 10e7

                time.sleep(wait)

    except KeyboardInterrupt:  # pragma: no cover
        exit_handler()


def kill_old_astrality_processes() -> None:
    """
    Kill any previous Astrality process instance.

    This process kills the last process which invoked this function.
    If the process is no longer running, it is owned by another user, or has
    a new create_time, it will *not* be killed.
    """
    # The current process
    new_process = psutil.Process()

    # Fetch info of possible previous process instance
    pidfile = XDG().data('astrality.pid')
    old_process_info = utils.load_yaml(path=pidfile)
    utils.dump_yaml(
        data=new_process.as_dict(attrs=['pid', 'create_time', 'username']),
        path=pidfile,
    )

    if not old_process_info or not psutil.pid_exists(old_process_info['pid']):
        return

    try:
        old_process = psutil.Process(pid=old_process_info['pid'])
    except BaseException:
        return

    if not old_process.as_dict(
        attrs=['pid', 'create_time', 'username'],
    ) == old_process_info:
        return

    try:
        logger.info(
            'Killing duplicate Astrality process with pid: '
            f'{old_process.pid}.',
        )
        old_process.terminate()
        old_process.wait()
    except BaseException:
        logger.error(
            f'Could not kill old instance of astrality with pid: '
            f'{old_process.pid}. Continuing anyway...',
        )


if __name__ == '__main__':
    main()



"""Tests for the solar event listener subclass."""
from datetime import datetime, timedelta

from dateutil.tz import tzlocal
import pytest

from astrality.event_listener import Solar


@pytest.fixture
def solar_config():
    """A solar event listener configuration in Trondheim, Norway."""
    return {
        'type': 'solar',
        'latitude': 0,
        'longitude': 0,
        'elevation': 0,
    }


@pytest.fixture
def solar(solar_config):
    """A solar event listener in Trondheim, Norway."""
    return Solar(solar_config)


# --- Times around dawn ---


@pytest.fixture
def dawn(solar):
    return solar.construct_astral_location().sun()['dawn']


@pytest.fixture
def before_dawn(dawn):
    delta = timedelta(minutes=-2)
    return dawn + delta


@pytest.fixture
def after_dawn(dawn):
    delta = timedelta(minutes=2)
    return dawn + delta


def test_that_night_is_correctly_identified(solar, before_dawn, freezer):
    freezer.move_to(before_dawn)
    event = solar.event()
    assert event == 'night'


def test_that_sunrise_is_correctly_identified(solar, after_dawn, freezer):
    freezer.move_to(after_dawn)
    event = solar.event()
    assert event == 'sunrise'


# --- Times around dusk ---
@pytest.fixture
def dusk(solar):
    return solar.construct_astral_location().sun()['dusk']


@pytest.fixture
def before_dusk(dusk):
    delta = timedelta(minutes=-2)
    return dusk + delta


@pytest.fixture
def after_dusk(dusk):
    delta = timedelta(minutes=2)
    return dusk + delta


def test_that_night_is_correctly_identified_after_dusk(
    solar,
    after_dusk,
    freezer,
):
    freezer.move_to(after_dusk)
    event = solar.event()
    assert event == 'night'


def test_that_sunset_is_correctly_identified_before_dusk(
    solar,
    before_dusk,
    freezer,
):
    freezer.move_to(before_dusk)
    event = solar.event()
    assert event == 'sunset'


def test_location(solar):
    location = solar.construct_astral_location()
    assert str(location) \
        == 'CityNotImportant/RegionIsNotImportantEither, '\
           'tz=UTC, lat=0.00, lon=0.00'


def test_time_left_before_new_event(solar, before_dusk, freezer):
    freezer.move_to(before_dusk)
    assert solar.time_until_next_event().total_seconds() == 120


def test_time_right_before_midnight(solar, freezer):
    """
    This function requires special handling when the UTC time is later than all
    solar events within the same day, which is the case right before midnight.
    """

    before_midnight = datetime(
        year=2019,
        month=12,
        day=23,
        hour=23,
        second=59,
        microsecond=0,
    )
    freezer.move_to(before_midnight)

    # Test that the time left is within the bounds of 0 to 24 hours
    time_left = solar.time_until_next_event()
    assert 0 < time_left.total_seconds() < 60 * 60 * 24


def test_config_event_listener_method():
    solar_event_listener_application_config = {'type': 'solar'}
    solar_event_listener = Solar(solar_event_listener_application_config)
    assert solar_event_listener.event_listener_config['latitude'] == 0


@pytest.mark.parametrize(
    'hour,sun',
    [
        (23, 'night'),
        (1, 'night'),
        (5, 'sunrise'),
        (10, 'morning'),
        (13, 'afternoon'),
        (22, 'sunset'),
    ],
)
def test_locations_where_some_events_never_occur(freezer, hour, sun):
    """
    Test that locations with missing solar events are handled gracefully.

    During summer, closer to the poles, the sun never dips properly below
    the horizon. In this case astral throws an AstralError, and we have
    to fall back to some hard coded defaults instead.
    """
    summer = datetime(
        year=2018,
        month=5,
        day=24,
        hour=hour,
        minute=0,
        tzinfo=tzlocal(),
    )
    freezer.move_to(summer)

    polar_location = {
        'type': 'solar',
        'latitude': 89,
        'longitude': 89,
        'elevation': 0,
    }
    polar_sun = Solar(polar_location)
    assert polar_sun.event() == sun
    assert 0 < polar_sun.time_until_next_event().total_seconds() < 24 * 60 * 60



"""Tests for locating and parsing YAML configuration files."""
import os
from pathlib import Path
from shutil import rmtree

import pytest

from astrality.config import (
    ASTRALITY_DEFAULT_GLOBAL_SETTINGS,
    create_config_directory,
    infer_config_location,
    resolve_config_directory,
    user_configuration,
)
from astrality.context import Context
from astrality.utils import compile_yaml, dump_yaml


@pytest.fixture
def dummy_config():
    """Return dummy configuration YAML file."""
    test_conf = Path(__file__).parents[1] / 'test_config' / 'test.yml'
    return compile_yaml(
        path=test_conf,
        context=Context(),
    )


class TestAllConfigFeaturesFromDummyConfig:
    """Tests for .utils.compile_yaml."""

    def test_normal_variable(self, dummy_config):
        """String literals should be interpreted without modifications."""
        assert dummy_config['section1']['var1'] == 'value1'
        assert dummy_config['section1']['var2'] == 'value1/value2'
        assert dummy_config['section2']['var3'] == 'value1'

    def test_empty_string_variable(self, dummy_config):
        """Empty strings  should be representable."""
        assert dummy_config['section2']['empty_string_var'] == ''

    def test_non_existing_variable(self, dummy_config):
        """Non-existing keys should not have defaults but raise instead."""
        with pytest.raises(KeyError):
            assert dummy_config['section2']['not_existing_option'] \
                is None

    def test_environment_variable_interpolation(self, dummy_config):
        """Environment variables should be interpolated with Jinja syntax."""
        assert dummy_config['section3']['env_variable'] \
            == 'test_value, hello'


class TestResolveConfigDirectory:
    """Tests for .config.resolve_config_directory."""

    def test_setting_directory_using_application_env_variable(
        self,
        monkeypatch,
    ):
        """ASTRALITY_CONFIG_HOME should override XDG_CONFIG_HOME."""
        monkeypatch.setattr(
            os,
            'environ',
            {
                'ASTRALITY_CONFIG_HOME': '/test/dir',
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/test/dir')

    def test_setting_directory_using_xdg_directory_standard(self, monkeypatch):
        """The XDG_CONFIG_HOME environment variable should be respected."""
        monkeypatch.setattr(
            os,
            'environ',
            {
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/xdg/dir/astrality')

    def test_using_standard_config_dir_when_nothing_else_is_specified(
        self,
        monkeypatch,
    ):
        """In the absence of XDG_CONFIG_HOME, the standard location is used."""
        monkeypatch.setattr(os, 'environ', {})
        assert resolve_config_directory() \
            == Path('~/.config/astrality').expanduser()


class TestCreateConfigDirectory:
    """Tests for creating configuration content, used be the CLI entrypoint."""

    def test_creation_of_empty_config_directory(self):
        """An empty configuration directory can be created."""
        config_path = Path('/tmp/config_test')
        config_dir = create_config_directory(path=config_path, empty=True)

        assert config_path == config_dir
        assert config_dir.is_dir()
        assert len(list(config_dir.iterdir())) == 0

        config_dir.rmdir()

    def test_creation_of_infered_config_directory(self, monkeypatch):
        """When no directory is specified, it is inferred instead."""
        config_path = Path('/tmp/astrality_config')
        monkeypatch.setattr(
            os,
            'environ',
            {'ASTRALITY_CONFIG_HOME': str(config_path)},
        )
        created_config_dir = create_config_directory(empty=True)
        assert created_config_dir == config_path
        created_config_dir.rmdir()

    def test_creation_of_config_directory_with_example_content(self):
        """Test copying example configuration contents."""
        config_path = Path('/tmp/astrality_config_with_contents')
        created_config_dir = create_config_directory(config_path)
        assert created_config_dir == config_path

        # Test presence of content in created folder
        dir_contents = tuple(file.name for file in created_config_dir.iterdir())
        assert 'astrality.yml' in dir_contents
        assert 'modules' in dir_contents
        rmtree(created_config_dir)


class TestInferConfigLocation:
    """Tests for config.infer_config_location()."""

    def test_that_empty_config_location_still_return_paths(
        self,
        monkeypatch,
        caplog,
    ):
        """
        A lack of astrality.yml should not change the result.

        When astrality.yml is not found in the configuration directory, a
        warning should be logged, but the path should still be returned.
        The remaining logic should use default values for astrality.yml instead.
        """
        config_path = Path('/tmp/astrality_config')
        monkeypatch.setattr(
            os,
            'environ',
            {'ASTRALITY_CONFIG_HOME': str(config_path)},
        )
        caplog.clear()
        directory, config_file = infer_config_location()
        assert directory == config_path
        assert config_file == config_path / 'astrality.yml'
        assert 'not found' in caplog.record_tuples[0][2]


class TestUserConfiguration:
    """Tests for config.user_configuration()."""

    def test_missing_global_configuration_file(self, monkeypatch, tmpdir):
        """Missing astrality.yml should result in default values."""
        # Create directory used as astrality config directory
        config_home = Path(tmpdir)
        monkeypatch.setattr(
            os,
            'environ',
            {'ASTRALITY_CONFIG_HOME': str(config_home)},
        )

        # Sanity check
        assert len(list(config_home.iterdir())) == 0

        # Create modules and context files, but *not* astrality.yml
        modules = {'A': {'enabled': False}}
        dump_yaml(path=config_home / 'modules.yml', data=modules)

        context = {'section': {'key': 'value'}}
        dump_yaml(path=config_home / 'context.yml', data=context)

        (
            global_config,
            global_modules,
            global_context,
            inferred_path,
        ) = user_configuration()
        assert global_config == ASTRALITY_DEFAULT_GLOBAL_SETTINGS
        assert global_modules == modules
        assert global_context == context
        assert inferred_path == config_home



"""Tests for module manager behaviour related to file system modifications."""
import os
import shutil
from pathlib import Path
from sys import platform

import pytest

from astrality import utils
from astrality.context import Context
from astrality.module import ModuleManager
from astrality.tests.utils import Retry

MACOS = platform == 'darwin'


@pytest.yield_fixture
def modules_config(test_config_directory, temp_directory):
    empty_template = test_config_directory / 'templates' / 'empty.template'
    empty_template_target = empty_template.parent / 'empty_temp_template'
    touch_target = temp_directory / 'touched'

    secondary_template = test_config_directory \
        / 'templates' / 'no_context.template'
    secondary_template_target = temp_directory / 'secondary_template.tmp'

    modules = {
        'A': {
            'on_modified': {
                str(empty_template): {
                    'compile': [
                        {
                            'content': str(empty_template),
                            'target': str(empty_template_target),
                        },
                        {
                            'content': str(secondary_template),
                            'target': str(secondary_template_target),
                        },
                    ],
                    'run': [{'shell': 'touch ' + str(touch_target)}],
                },
            },
        },
        'B': {},
    }

    yield (
        modules,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    )

    # Cleanup all test files for next test iteration
    if empty_template.is_file():
        empty_template.write_text('')
    if empty_template_target.is_file():
        os.remove(empty_template_target)
    if secondary_template_target.is_file():
        os.remove(secondary_template_target)
    if touch_target.is_file():
        os.remove(touch_target)


def test_modified_commands_of_module(modules_config):
    modules, empty_template, empty_template_target, touch_target, *_ \
        = modules_config
    module_manager = ModuleManager(
        modules=modules,
    )
    result = module_manager.modules['A'].execute(
        action='run',
        block='on_modified',
        path=empty_template,
    )
    assert result == (('touch ' + str(touch_target), ''),)


def test_direct_invocation_of_modifed_method_of_module_manager(modules_config):
    (
        modules,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    ) = modules_config

    module_manager = ModuleManager(modules=modules)

    # PS: Disabling the directory watcher is not necessary, as it is done in
    # the startup method.

    # Now write new text to the template
    empty_template.write_text('new content')

    # And trigger the modified method manually
    module_manager.file_system_modified(empty_template)

    # And assert that the new template has been compiled
    assert empty_template_target.is_file()
    with open(empty_template_target) as file:
        assert file.read() == 'new content'

    # And that the new file has been touched
    assert Retry()(lambda: touch_target.is_file())


@pytest.mark.slow
@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
def test_on_modified_event_in_module(modules_config):
    (
        modules,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    ) = modules_config

    module_manager = ModuleManager(modules=modules)

    # Start the file watcher by invoking the startup command indirectly
    # through finish_tasks() method
    module_manager.finish_tasks()

    # Assert that the template file is really empty as a sanity check
    assert empty_template.read_text() == ''

    # And that target files are not created yet
    assert not touch_target.is_file()
    assert not empty_template_target.is_file()
    assert not secondary_template_target.is_file()

    # Trigger the on_modified event
    empty_template.write_text('new content')

    # And assert that the new template has been compiled
    retry = Retry()
    assert retry(lambda: empty_template_target.is_file())
    assert retry(lambda: empty_template_target.read_text() == 'new content')

    # Assert that also templates from other modules are compiled
    assert retry(lambda: secondary_template_target.is_file())
    assert retry(
        lambda: secondary_template_target.read_text() == 'one\ntwo\nthree',
    )

    # And that the new file has been touched
    assert retry(lambda: touch_target.is_file())


@pytest.yield_fixture
def test_template_targets():
    template_target1 = Path('/tmp/astrality/target1')
    template_target2 = Path('/tmp/astrality/target2')

    yield template_target1, template_target2

    if template_target1.is_file():
        os.remove(template_target1)
    if template_target2.is_file():
        os.remove(template_target2)


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_hot_reloading(
    test_template_targets,
    test_config_directory,
):
    template_target1, template_target2 = test_template_targets
    config1 = test_config_directory / 'modules1.yml'
    config2 = test_config_directory / 'modules2.yml'
    target_config = test_config_directory / 'modules.yml'

    # Copy the first configuration into place
    shutil.copy(str(config1), str(target_config))

    modules1 = utils.compile_yaml(
        config1,
        context={},
    )

    application_config1 = {'astrality': {'hot_reload_config': True}}
    module_manager = ModuleManager(
        config=application_config1,
        modules=modules1,
        directory=test_config_directory,
    )

    # Before beginning, the template should not be compiled
    assert not template_target1.is_file()

    # But when we finalize tasks, it should be compiled
    module_manager.finish_tasks()
    assert template_target1.is_file()

    # Also check that the filewatcher has been started
    assert module_manager.directory_watcher.observer.is_alive()

    # We now "edit" the configuration file
    shutil.copy(str(config2), str(target_config))

    # Since hot reloading is enabled, the new template target should be
    # compiled, and the old one cleaned up
    retry = Retry()
    assert retry(lambda: template_target2.is_file())
    assert retry(lambda: not template_target1.is_file())

    # And we switch back again
    shutil.copy(str(config1), str(target_config))
    assert retry(lambda: template_target1.is_file())
    assert retry(lambda: not template_target2.is_file())

    # Cleanup config file
    if target_config.is_file():
        os.remove(target_config)

    # Stop the filewatcher
    module_manager.directory_watcher.stop()


@pytest.yield_fixture
def three_watchable_files(test_config_directory):
    file1 = test_config_directory / 'file1.tmp'
    file2 = test_config_directory / 'file2.tmp'
    file3 = test_config_directory / 'file3.tmp'

    # Delete files if they are present
    if file1.is_file():
        os.remove(file1)
    if file2.is_file():
        os.remove(file2)
    if file3.is_file():
        os.remove(file3)

    yield file1, file2, file3

    # Delete files on cleanup
    if file1.is_file():
        os.remove(file1)
    if file2.is_file():
        os.remove(file2)
    if file3.is_file():
        os.remove(file3)


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_all_three_actions_in_on_modified_block(
    three_watchable_files,
    test_config_directory,
):
    file1, file2, file3 = three_watchable_files
    car_template = test_config_directory / 'templates' / 'a_car.template'
    mercedes_context = test_config_directory / 'context' / 'mercedes.yml'
    tesla_context = test_config_directory / 'context' / 'tesla.yml'

    modules = {
        'car': {
            'on_startup': {
                'import_context': {
                    'from_path': str(mercedes_context),
                },
                'compile': {
                    'content': str(car_template),
                    'target': str(file1),
                },
            },
            'on_modified': {
                str(file2): {
                    'import_context': {
                        'from_path': str(tesla_context),
                    },
                    'compile': {
                        'content': str(car_template),
                        'target': str(file1),
                    },
                    'run': {'shell': 'touch ' + str(file3)},
                },
            },
        },
    }
    module_manager = ModuleManager(modules=modules)

    # Sanity check before beginning testing
    assert not file1.is_file()
    assert not file2.is_file()
    assert not file3.is_file()

    # Now finish tasks, i.e. on_startup block
    module_manager.finish_tasks()
    assert file1.is_file()
    assert not file2.is_file()
    assert not file3.is_file()

    # Check that the correct context is inserted
    with open(file1) as file:
        assert file.read() == 'My car is a Mercedes'

    # Now modify file2 such that the on_modified block is triggered
    file2.write_text('some new content')

    # The on_modified run command should now have been executed
    assert Retry()(lambda: file3.is_file())

    module_manager.exit()


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_recompile_templates_when_modified(three_watchable_files):
    template, target, _ = three_watchable_files
    template.touch()

    modules = {
        'module_name': {
            'on_startup': {
                'compile': {
                    'content': str(template),
                    'target': str(target),
                },
            },
        },
    }

    application_config = {'modules': {'reprocess_modified_files': True}}
    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
        context=Context({
            'section': {1: 'value'},
        }),
    )

    # Sanity check before beginning testing
    with open(template) as file:
        assert file.read() == ''

    assert not target.is_file()

    # Compile the template
    module_manager.finish_tasks()
    with open(target) as file:
        assert file.read() == ''

    # Now write to the template and see if it is recompiled
    template.write_text('{{ section.2 }}')
    assert Retry()(lambda: target.read_text() == 'value')

    module_manager.exit()
    module_manager.directory_watcher.stop()


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_recompile_templates_when_modified_overridden(
    three_watchable_files,
    test_config_directory,
):
    """
    If a file is watched in a on_modified block, it should override the
    reprocess_modified_files option.
    """
    template, target, touch_target = three_watchable_files
    template.touch()

    modules = {
        'module_name': {
            'on_startup': {
                'compile': {
                    'content': str(template),
                    'target': str(target),
                },
            },
            'on_modified': {
                str(template): {
                    'run': {'shell': 'touch ' + str(touch_target)},
                },
            },
        },
    }

    application_config = {'modules': {'reprocess_modified_files': True}}
    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
        context=Context({
            'section': {1: 'value'},
        }),
        directory=test_config_directory,
    )

    # Sanity check before beginning testing
    with open(template) as file:
        assert file.read() == ''

    assert not target.is_file()

    # Compile the template
    module_manager.finish_tasks()
    with open(target) as file:
        assert file.read() == ''

    # Now write to the template and see if it is *compiled*, but the on_modified
    # command is run instead
    template.write_text('{{ section.2 }}')

    retry = Retry()
    assert retry(lambda: target.read_text() == '')
    assert retry(lambda: touch_target.is_file())

    module_manager.exit()


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_importing_context_on_modification(
    three_watchable_files,
    test_config_directory,
):
    """Test that context values are imported in on_modified blocks."""
    file1, *_ = three_watchable_files
    mercedes_context = test_config_directory / 'context' / 'mercedes.yml'

    modules = {
        'module_name': {
            'on_modified': {
                str(file1): {
                    'import_context': {
                        'from_path': str(mercedes_context),
                    },
                },
            },
        },
    }
    module_manager = ModuleManager(
        modules=modules,
        context=Context({
            'car': {'manufacturer': 'Tesla'},
        }),
    )
    module_manager.finish_tasks()

    # Sanity check before modifying file1
    assert module_manager.application_context['car']['manufacturer'] == 'Tesla'

    # After modifying file1, Mercedes should have been imported
    file1.touch()
    file1.write_text('new content, resulting in importing Mercedes')
    assert Retry()(
        lambda: module_manager.application_context
        ['car']['manufacturer'] == 'Mercedes',
    )


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_that_stowed_templates_are_also_watched(three_watchable_files):
    """Stowing template instead of compiling it should still be watched."""
    template, target, _ = three_watchable_files
    template.touch()

    modules = {
        'module_name': {
            'on_startup': {
                'stow': {
                    'content': str(template),
                    'target': str(target),
                    'templates': '(.+)',
                    'non_templates': 'ignore',
                },
            },
        },
    }

    application_config = {'modules': {'reprocess_modified_files': True}}
    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
        context=Context({
            'section': {1: 'value'},
        }),
    )

    # Sanity check before beginning testing
    with open(template) as file:
        assert file.read() == ''

    assert not target.is_file()

    # Stow the template
    module_manager.finish_tasks()
    with open(target) as file:
        assert file.read() == ''

    # Now write to the template and see if it is recompiled
    template.write_text('{{ section.2 }}')
    assert Retry()(lambda: target.read_text() == 'value')

    module_manager.exit()



"""Tests for ensuring that all files that are created are persisted."""

import pytest

from astrality.module import ModuleManager
from astrality.persistence import CreatedFiles


@pytest.mark.parametrize('method', ['compile', 'copy', 'symlink'])
def test_that_created_files_are_persisted(method, create_temp_files):
    """When modules create files, they should be persisted."""
    (
        template1,
        template2,
        template3,
        target1,
        target2,
        target3,
    ) = create_temp_files(6)

    # Delete targets to prevent backups from being restored
    for target in (target1, target2, target3):
        target.unlink()

    modules = {
        'A': {
            method: [
                {
                    'content': str(template1),
                    'target': str(target1),
                },
                {
                    'content': str(template2),
                    'target': str(target2),
                },
            ],
        },
        'B': {
            method: {
                'content': str(template3),
                'target': str(target3),
            },
        },
    }
    module_manager = ModuleManager(modules=modules)
    module_manager.finish_tasks()

    created_files = CreatedFiles()
    assert created_files.by(module='A') == [target1, target2]
    assert created_files.by(module='B') == [target3]

    # Now we should be able to cleanup the created files
    assert target1.exists()
    assert target2.exists()
    assert target3.exists()

    # First let's see if dry run is respected
    created_files.cleanup(module='A', dry_run=True)
    assert target1.exists()
    assert target2.exists()
    assert target3.exists()
    assert created_files.by(module='A') == [target1, target2]
    assert created_files.by(module='B') == [target3]

    # Now see if we can cleanup module A and let B stay intact
    created_files.cleanup(module='A')
    assert not target1.exists()
    assert not target2.exists()
    assert target3.exists()
    assert created_files.by(module='A') == []
    assert created_files.by(module='B') == [target3]

    # Now all files should be cleaned
    created_files.cleanup(module='B')
    assert not target3.exists()
    assert created_files.by(module='A') == []
    assert created_files.by(module='B') == []

    # Let's see if it has been properly persisted too
    del created_files
    created_files = CreatedFiles()
    assert created_files.by(module='A') == []
    assert created_files.by(module='B') == []



"""Tests for utils.resolve_targets."""

from pathlib import Path

from astrality.utils import resolve_targets


def test_resolving_non_existent_file():
    """When `content` does not exist, no targets should be returned."""
    targets = resolve_targets(
        content=Path('/does/not/exist'),
        target=Path('/'),
        include=r'.*',
    )
    assert targets == {}


def test_resolving_target_of_content_file():
    """When `content` is a file, the target root is used."""
    targets = resolve_targets(
        content=Path(__file__),
        target=Path('/does/not/exist'),
        include=r'.*',
    )
    assert targets == {Path(__file__): Path('/does/not/exist')}


def test_resolving_target_file_to_directory():
    """When content is a file, but target is a directory, keep filename."""
    targets = resolve_targets(
        content=Path(__file__),
        target=Path('/tmp'),
        include=r'.*',
    )
    assert targets == {Path(__file__): Path('/tmp/test_resolve_targets.py')}


def test_resolving_content_directory(tmpdir):
    """Directory hierarchy should be preserved at target."""
    temp_dir = Path(tmpdir)

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    targets = resolve_targets(
        content=temp_dir,
        target=Path('/a/b'),
        include=r'.*',
    )
    assert targets == {
        file1: Path('/a/b/file1'),
        file2: Path('/a/b/file2'),
        file3: Path('/a/b/recursive/file3'),
    }


def test_filtering_based_on_include(tmpdir):
    """Only files that match the regex should be included."""
    temp_dir = Path(tmpdir)

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    targets = resolve_targets(
        content=temp_dir,
        target=Path('/a/b'),
        include=r'.+3',
    )
    assert targets == {
        file3: Path('/a/b/recursive/file3'),
    }


def test_renaming_based_on_include(tmpdir):
    """Targets should be renameable based on the include capture group."""
    temp_dir = Path(tmpdir)

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    targets = resolve_targets(
        content=temp_dir,
        target=Path('/a/b'),
        include=r'.+(\d)',
    )
    assert targets == {
        file1: Path('/a/b/1'),
        file2: Path('/a/b/2'),
        file3: Path('/a/b/recursive/3'),
    }



"""Tests for astrality.actions.StowAction."""

from pathlib import Path

from astrality.actions import StowAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Copy actions without options should do nothing."""
    stow_action = StowAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    stow_action.execute()


def test_filtering_stowed_templates(test_config_directory, tmpdir):
    """Users should be able to restrict compilable templates with ignore."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'
    stow_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'templates': r'.+\.template',
        'non_templates': 'ignore',
    }
    stow_action = StowAction(
        options=stow_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # First testing if dry run is respected (too much work for a separate test)
    stow_action.execute(dry_run=True)
    assert len(list(temp_dir.iterdir())) == 0

    # We should have a total of two stowed files
    stow_action.execute()
    assert len(list(temp_dir.iterdir())) == 2
    assert len(list((temp_dir / 'recursive').iterdir())) == 1
    assert (temp_dir / 'module.template').is_file()
    assert (temp_dir / 'recursive' / 'empty.template').is_file()


def test_renaming_templates(test_config_directory, tmpdir):
    """Templates targets should be renameable with a capture group."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'

    # Multiple capture groups should be allowed
    stow_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'templates': r'(?:^template\.(.+)$|^(.+)\.template$)',
        'non_templates': 'ignore',
    }
    stow_action = StowAction(
        options=stow_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    stow_action.execute()

    # We should have a total of two stowed files
    assert len(list(temp_dir.iterdir())) == 2
    assert len(list((temp_dir / 'recursive').iterdir())) == 1
    assert (temp_dir / 'module').is_file()
    assert (temp_dir / 'recursive' / 'empty').is_file()


def test_symlinking_non_templates(test_config_directory, tmpdir):
    """Non-templates files should be implicitly symlinked."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'
    stow_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'templates': r'.+\.template',
    }
    stow_action = StowAction(
        options=stow_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    stow_action.execute()

    # Templates should still stowed
    target_dir_content = list(temp_dir.iterdir())
    assert len(target_dir_content) == 6
    assert temp_dir / 'module.template' in target_dir_content
    assert not (temp_dir / 'module.template').is_symlink()
    assert (temp_dir / 'recursive' / 'empty.template').is_file()

    # The rest should be symlinked
    assert (temp_dir / 'modules.yml').is_symlink()
    assert (temp_dir / 'modules.yml').resolve() == templates / 'modules.yml'

    # Symlinked files should be not considered as a managed file, as it is
    # self-updating.
    assert templates / 'modules.yml' not in stow_action.managed_files()


def test_copying_non_template_files(test_config_directory, tmpdir):
    """Non-templates files can be copied."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'
    stow_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'templates': r'.+\.template',
        'non_templates': 'copy',
    }
    stow_action = StowAction(
        options=stow_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    stow_action.execute()

    # Templates should still stowed
    target_dir_content = list(temp_dir.iterdir())
    assert len(target_dir_content) == 6
    assert temp_dir / 'module.template' in target_dir_content
    assert (temp_dir / 'recursive' / 'empty.template').is_file()

    # The rest should be copied
    assert (temp_dir / 'modules.yml').is_file()
    assert (temp_dir / 'modules.yml').read_text() == \
        (templates / 'modules.yml').read_text()

    # Copied files should be considered as a managed file, as it needs to be
    # copied again if modified.
    assert templates / 'modules.yml' in stow_action.managed_files()



#!/usr/bin/env python3.6
"""Distribution and installation of Astrality."""

from pathlib import Path

from setuptools import find_packages, setup


def readme():
    with open(Path(__file__).parent / 'README.rst') as file:
        return file.read()


setup(
    name='astrality',
    version='1.1.1',
    packages=find_packages(),
    install_requires=[
        'Jinja2',
        'astral',
        'coloredlogs',
        'mypy_extensions',
        'psutil',
        'python-dateutil',
        'pyyaml',
        'requests',
        'watchdog',
    ],
    python_requires='>=3.6',
    scripts=['bin/astrality'],
    include_package_data=True,

    # metadata for upload to PyPI
    author='Jakob Gerhard Martinussen',
    author_email='jakobgm@gmail.com',
    description='A dynamic configuration file manager.',
    long_description=readme(),
    license='MIT',
    keywords='unix configuration management',
    url='http://github.com/JakobGM/astrality',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.6',
        'Topic :: Desktop Environment :: File Managers',
        'Topic :: Text Processing',
    ],
)



"""Tests for configuraition dictionary management."""
import pytest

from astrality.module import GlobalModulesConfig, ModuleManager


def test_modules_config_explicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'burma::burma'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    for source in global_modules_config.external_module_sources:
        source.modules({})

    assert 'burma::burma' in global_modules_config.enabled_modules
    assert 'india' in global_modules_config.enabled_modules
    assert 'whatever' not in global_modules_config.enabled_modules


def test_modules_config_implicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'burma::*'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    for source in global_modules_config.external_module_sources:
        source.modules({})

    assert 'burma::burma' in global_modules_config.enabled_modules
    assert 'india' in global_modules_config.enabled_modules
    assert 'burma::only_defined_accepted' \
        not in global_modules_config.enabled_modules


def test_modules_config_several_implicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'two_modules::*'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    for source in global_modules_config.external_module_sources:
        source.modules({})

    assert 'two_modules::bhutan' in global_modules_config.enabled_modules
    assert 'two_modules::bangladesh' in global_modules_config.enabled_modules


def test_modules_config_where_all_modules_are_enabled(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': str(test_config_directory / 'test_modules'),
        'enabled_modules': [
            {'name': '*::*'},
            {'name': '*'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    global_modules_config.compile_config_files(
        {'module': {'setting': 'whatever'}},
    )

    assert 'two_modules::bhutan' in global_modules_config.enabled_modules
    assert 'two_modules::bangladesh' in global_modules_config.enabled_modules


def test_enabling_of_modules_defined_different_places():
    application_config = {
        'modules': {
            'modules_directory': 'freezed_modules',
            'enabled_modules': [
                {'name': 'south_america::brazil'},
                {'name': 'india'},
            ],
        },
    }
    modules = {
        'india': {},     # Enabled
        'pakistan': {},  # Not enabled
    }
    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
    )

    assert len(module_manager.modules) == 2
    assert 'south_america::brazil' in module_manager.modules
    assert 'india' in module_manager.modules


def test_enabling_of_all_modules():
    application_config = {
        'modules': {
            'modules_directory': 'freezed_modules',
        },
    }

    modules = {
        'india': {},     # Enabled
        'pakistan': {},  # Not enabled
    }

    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
    )

    assert len(module_manager.modules) == 5
    assert 'india' in module_manager.modules
    assert 'pakistan' in module_manager.modules
    assert 'north_america::USA' in module_manager.modules
    assert 'south_america::brazil' in module_manager.modules
    assert 'south_america::argentina' in module_manager.modules


@pytest.mark.slow
def test_using_three_different_module_sources(test_config_directory):
    modules_directory = test_config_directory / 'freezed_modules'

    application_config = {
        'modules': {
            'modules_directory': str(modules_directory),
            'enabled_modules': [
                {'name': 'north_america::*'},
                {'name': 'github::jakobgm/test-module.astrality'},
                {'name': 'italy'},
            ],
        },
    }

    modules = {
        'italy': {},
        'spain': {},
    }
    module_manager = ModuleManager(
        config=application_config,
        modules=modules,
    )

    assert len(module_manager.modules) == 4
    assert 'north_america::USA' in module_manager.modules
    assert 'github::jakobgm/test-module.astrality::botswana' \
        in module_manager.modules
    assert 'github::jakobgm/test-module.astrality::ghana' \
        in module_manager.modules
    assert 'italy' in module_manager.modules



"""Tests for the --cleanup cli flag."""

import os
from pathlib import Path

import pytest

from astrality.module import ModuleManager


@pytest.mark.parametrize('method', ['compile', 'copy', 'symlink'])
def test_that_cleanup_cli_works(
    method,
    create_temp_files,
    patch_xdg_directory_standard,
):
    """--cleanup module_name, all module created files should be deleted"""
    (
        template1,
        template2,
        template3,
        target1,
        target2,
        target3,
    ) = create_temp_files(6)

    for template in (template1, template2, template3):
        template.write_text('new content')

    for target in (target1, target2, target3):
        target.write_text('original content')

    modules = {
        'A': {
            method: [
                {
                    'content': str(template1),
                    'target': str(target1),
                },
                {
                    'content': str(template2),
                    'target': str(target2),
                },
            ],
        },
        'B': {
            method: {
                'content': str(template3),
                'target': str(target3),
            },
        },
    }
    module_manager = ModuleManager(modules=modules)
    module_manager.finish_tasks()

    for target in (target1, target2, target3):
        assert target.resolve().read_text() == 'new content'

    bin_script = str(Path(__file__).parents[3] / 'bin' / 'astrality')
    data_home = 'XDG_DATA_HOME="' + str(patch_xdg_directory_standard) + '/.." '
    command = data_home + bin_script + ' --cleanup A --cleanup B'
    os.system(command)

    for target in (target1, target2, target3):
        assert target.resolve().read_text() == 'original content'



"""Tests for astrality.actions.CopyAction."""

from pathlib import Path

from astrality.actions import CopyAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Copy actions without options should do nothing."""
    copy_action = CopyAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    copy_action.execute()


def test_if_dry_run_is_respected(create_temp_files, caplog):
    """When dry_run is True, the copy action should only be logged."""
    content, target = create_temp_files(2)
    content.write_text('content')
    target.write_text('target')

    copy_action = CopyAction(
        options={'content': str(content), 'target': str(target)},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    result = copy_action.execute(dry_run=True)

    # We should still return the copy pair
    assert result == {content: target}

    # We should log what would have been done
    assert 'SKIPPED:' in caplog.record_tuples[0][2]
    assert str(content) in caplog.record_tuples[0][2]
    assert str(target) in caplog.record_tuples[0][2]

    # But we should not copy the file under a dry run
    assert target.read_text() == 'target'


def test_copy_action_using_all_parameters(tmpdir):
    """All three parameters should be respected."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.write_text('file1 content')

    file2 = temp_dir / 'file2'
    file2.write_text('file2 content')

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.write_text('file3 content')

    copy_options = {
        'content': str(temp_dir),
        'target': str(target),
        'include': r'file(\d)',
    }
    copy_action = CopyAction(
        options=copy_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    copy_action.execute()

    assert (target / '1').read_text() == file1.read_text()
    assert (target / '2').read_text() == file2.read_text()
    assert (target / 'recursive' / '3').read_text() == file3.read_text()
    assert copy_action.copied_files == {
        file1: {target / '1'},
        file2: {target / '2'},
        file3: {target / 'recursive' / '3'},
    }
    assert file1 in copy_action


def test_copying_without_renaming(tmpdir):
    """When include is not given, keep copy name."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    copy_options = {
        'content': str(temp_dir),
        'target': str(target),
    }
    copy_action = CopyAction(
        options=copy_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    copy_action.execute()

    assert (target / 'file1').read_text() == file1.read_text()
    assert (target / 'file2').read_text() == file2.read_text()
    assert (target / 'recursive' / 'file3').read_text() == file3.read_text()


def test_copying_file_to_directory(tmpdir):
    """If copying from directory to file, place file in directory."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()

    copy_options = {
        'content': str(file1),
        'target': str(target),
        'include': r'file1',
    }
    copy_action = CopyAction(
        options=copy_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    copy_action.execute()

    assert (target / 'file1').read_text() == file1.read_text()


def test_setting_permissions_on_target_copy(tmpdir):
    """If permissions is provided, use it for the target."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()
    file1.chmod(0o770)

    copy_options = {
        'content': str(file1),
        'target': str(target),
        'include': r'file1',
        'permissions': '777',
    }
    copy_action = CopyAction(
        options=copy_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    copy_action.execute()

    assert ((target / 'file1').stat().st_mode & 0o000777) == 0o777


def test_backup_of_copy_target(create_temp_files):
    """Overwritten copy targets should be backed up."""
    target, content = create_temp_files(2)

    # This file is the original and should be backed up
    target.write_text('original')

    # This is the new content copied to target
    content.write_text('new')

    copy_options = {
        'content': str(content.name),
        'target': str(target),
    }
    copy_action = CopyAction(
        options=copy_options,
        directory=content.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # We replace the content by executing the action
    copy_action.execute()
    assert target.read_text() == 'new'

    # And when cleaning up the module, the backup should be restored
    CreatedFiles().cleanup(module='test')
    assert target.read_text() == 'original'


def test_cleanup_of_created_directory(create_temp_files, tmpdir):
    """Created directories should be cleaned up."""
    tmpdir = Path(tmpdir)
    [content] = create_temp_files(1)

    # The target requires a new directory to be created
    directory = tmpdir / 'dir'
    target = directory / 'target.tmp'

    # Execute copy action
    copy_options = {
        'content': str(content.name),
        'target': str(target),
    }
    created_files = CreatedFiles().wrapper_for(module='test')
    copy_action = CopyAction(
        options=copy_options,
        directory=content.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=created_files,
    )
    copy_action.execute()

    # The directory should now be created and persisted
    assert directory.is_dir()
    assert directory in created_files.creation_store

    # And it should be deleted on cleanup
    created_files.creation_store.cleanup(module='test')
    assert not directory.is_dir()



"""Tests for the test utils module."""

from .utils import RegexCompare


def test_use_of_regex_compare():
    regex_compare = RegexCompare(r'test')

    assert regex_compare == 'test'
    assert 'test' == regex_compare

    assert not 'ttest' == regex_compare


def test_more_complicated_regex_comparisons():
    regex_compare = RegexCompare(r'\[Compiling\] .+test\.template"')

    assert regex_compare == '[Compiling] something "/a/b/c/test.template"'
    assert not regex_compare == '[Compiling] something "/a/b/c/test.templates"'



"""Tests for Module stow action."""

from pathlib import Path


def test_stowing(
    action_block_factory,
    create_temp_files,
    module_factory,
):
    """Module should stow properly."""
    template, target = create_temp_files(2)
    template.write_text('{{ env.EXAMPLE_ENV_VARIABLE }}')
    symlink_target = template.parent / 'symlink_me'
    symlink_target.touch()

    action_block = action_block_factory(
        stow={
            'content': str(template.parent),
            'target': str(target.parent),
            'templates': r'file(0).temp',
            'non_templates': 'symlink',
        },
    )
    module = module_factory(
        on_exit=action_block,
    )
    module.execute(action='all', block='on_exit')

    # Check if template has been compiled
    assert Path(target.parent / '0').read_text() == 'test_value'

    # Check if non_template has been symlinked
    assert (template.parent / 'symlink_me').resolve() == symlink_target



"""General utility functions which are used across the application."""

import logging
import re
import shutil
import subprocess
from functools import partial
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, TypeVar, Union

from yaml import dump, load  # noqa

from astrality import compiler
from astrality.context import Context


logger = logging.getLogger(__name__)

# Try to import PyYAML library for faster YAML parsing
try:
    from yaml import CLoader as Loader, CDumper as Dumper  # type: ignore
    logger.info('Using LibYAML bindings for faster .yml parsing.')
except ImportError:  # pragma: no cover
    from yaml import Loader, Dumper
    logger.warning(
        'LibYAML not installed.'
        'Using somewhat slower pure python implementation.',
    )


def run_shell(
    command: str,
    timeout: Union[int, float] = 2,
    fallback: Any = '',
    working_directory: Path = Path.home(),
    allow_error_codes: bool = False,
    log_success: bool = True,
) -> str:
    """
    Return the standard output of a shell command.

    If the shell command has a non-zero exit code or times out, the function
    returns the `fallback` argument instead of the standard output.

    :param command: Shell command to be executed.
    :param timeout: How long to wait for return of command.
    :param fallback: Default return value on timeout/error codes.
    :param working_directory: Command working directory.
    :param allow_error_codes: If error codes should return fallback.
    :param log_success: If successful commands stdout should be logged.
    :return: Stdout of command, or `fallback` on error/timeout.
    """
    process = subprocess.Popen(
        command,
        cwd=working_directory,
        shell=True,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        # We add just a small extra wait in case users specify 0 seconds,
        # in order to not print an error when a command is really quick.
        if timeout == 0:
            process.wait(timeout=0.1)
        else:
            process.wait(timeout=timeout)

        for error_line in process.stderr:
            logger.error(str(error_line))

        if process.returncode != 0 and not allow_error_codes:
            logger.error(
                f'Command "{command}" exited with non-zero return code: ' +
                str(process.returncode),
            )
            return fallback
        else:
            stdout = process.communicate()[0].strip('\n')
            if log_success and stdout:
                logger.info(stdout)
            return stdout

    except subprocess.TimeoutExpired:
        if timeout == 0:
            return fallback

        logger.warning(
            f'The command "{command}" used more than {timeout} seconds in '
            'order to finish. The exit code can not be verified. This might be '
            'intentional for background processes and daemons.',
        )
        return fallback


T = TypeVar('T')


def cast_to_list(content: Union[T, List[T]]) -> List[T]:
    """
    Cast content to a 1-item list containing content.

    If content already is a list, return content unaltered.
    """
    if not isinstance(content, list):
        return [content]
    else:
        return content


def resolve_targets(
    content: Path,
    target: Path,
    include: str,
) -> Dict[Path, Path]:
    """
    Return content/target file path pairs.

    If `content` is a directory, it returns content/target path pairs that
    reflects `content`s file hierarchy at the `target` directory.

    If `content` is a file, and `target` is an existing directory, the target
    will become `content` within the `target`.

    Content files that do not match the include regex string are discarded.
    Any capture group in the regex string is used for renaming the target.

    :param content: Source path, either file or directory.
    :param target: Target path, either file or directory.
    :param include: Regular expression string for filtering/renaming content.
    :return: Dictionary with content file keys and target file values.
    """
    targets: Dict[Path, Path] = {}
    if content.is_file():
        if target.is_dir():
            targets[content] = target / content.name
        else:
            targets[content] = target

    else:
        for file in content.glob('**/*'):
            if file.is_dir():
                continue
            target_file = target / file.relative_to(content)
            targets[file] = target_file

    include_pattern = re.compile(include)
    filtered_targets: Dict[Path, Path] = {}
    for content_file, target_file in targets.items():
        match = include_pattern.fullmatch(target_file.name)
        if not match:
            continue

        # If there is no group match, keep the name, else, use last group
        renamed_target_file = target_file.parent \
            / match.group(match.lastindex or 0)
        filtered_targets[content_file] = renamed_target_file

    return filtered_targets


def compile_yaml(
    path: Path,
    context: Context,
) -> Dict:
    """
    Return datastructure from compiled YAML jinja2 template.

    :param path: YAML template file path.
    :param context: Jinja2 context.
    """
    if not path.is_file():  # pragma: no cover
        error_msg = f'Could not load config file "{path}".'
        logger.critical(error_msg)
        raise FileNotFoundError(error_msg)

    config_string = compiler.compile_template_to_string(
        template=path,
        context=context,
        shell_command_working_directory=path.parent,
    )

    return load(StringIO(config_string), Loader=Loader)


def load_yaml(path: Path) -> Any:
    """
    Load content from YAML file.

    :param path: Path to YAML formatted file.
    :return: Contents of YAML file.
    """
    with open(path, 'r') as yaml_file:
        return load(yaml_file.read(), Loader=Loader)


def dump_yaml(path: Path, data: Dict) -> None:
    """
    Dump data to yaml file.

    :param path: Path to file to be created.
    :param data: Data to be dumped to file.
    """
    str_data = yaml_str(data)
    with open(path, 'w') as yaml_file:
        yaml_file.write(str_data)


def yaml_str(data: Any) -> str:
    """
    Return YAML string representation of data.

    :param data: Data to be converted to YAML string format.
    :return: YAML string representation of python data structure.
    """
    return dump(
        data,
        Dumper=Dumper,
        default_flow_style=False,
    )


def copy(
    source: Union[str, Path],
    destination: Union[str, Path],
    follow_symlinks=True,
) -> None:
    """
    Copy source path content to destination path.

    :param source: Path to content to be copied.
    :param destination: New path for content.
    :param follow_symlinks: If True, symlinks are resolved before copying.
    """
    shutil.copy2(
        src=str(source),
        dst=str(destination),
        follow_symlinks=follow_symlinks,
    )


def move(
    source: Union[str, Path],
    destination: Union[str, Path],
    follow_symlinks=True,
) -> None:
    """
    Move source path content to destination path.

    :param source: Path to content to be moved.
    :param destination: New path for content.
    :param follow_symlinks: If True, symlinks are resolved before moving.
    """
    shutil.move(
        src=str(source),
        dst=str(destination),
        copy_function=partial(
            copy,
            follow_symlinks=follow_symlinks,
        ),
    )



"""Tests for compile action class."""

import os
from pathlib import Path

from astrality.actions import CompileAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Compilation action with no parameters should be a null object."""
    compile_action = CompileAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    target = compile_action.execute()
    assert target == {}


def test_compilation_of_template_to_temporary_file(template_directory):
    """Compile template to temporary file in absence of `target`."""
    compile_dict = {
        'content': 'no_context.template',
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    compilations = compile_action.execute()

    template = template_directory / 'no_context.template'
    assert template in compilations
    assert compilations[template].read_text() == 'one\ntwo\nthree'


def test_that_dry_run_skips_compilation(template_directory, tmpdir, caplog):
    """If dry_run is True, skip compilation of template"""
    compilation_target = Path(tmpdir, 'target.tmp')
    template = template_directory / 'no_context.template'
    compile_dict = {
        'content': 'no_context.template',
        'target': str(compilation_target),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    compilations = compile_action.execute(dry_run=True)

    # Check that the "compilation" is actually logged
    assert 'SKIPPED:' in caplog.record_tuples[0][2]
    assert str(template) in caplog.record_tuples[0][2]
    assert str(compilation_target) in caplog.record_tuples[0][2]

    # The template should still be returned
    assert template in compilations

    # And the compilation pair should be persisted
    assert compile_action.performed_compilations() == {
        template: {compilation_target},
    }

    # But the file should not be compiled
    assert not compilations[template].exists()


def test_compilation_to_specific_absolute_file_path(template_directory, tmpdir):
    """
    Compile to specified absolute target path.

    The template is specified relatively.
    """
    target = Path(tmpdir) / 'target'
    compile_dict = {
        'content': 'no_context.template',
        'target': str(target),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    return_target = list(compile_action.execute().values())[0]

    assert return_target == target
    assert target.read_text() == 'one\ntwo\nthree'


def test_compilation_to_specific_relative_file_path(template_directory, tmpdir):
    """
    Compile to specified absolute target path.

    The template is specified absolutely.
    """
    target = Path(tmpdir) / 'target'
    compile_dict = {
        'content': str(template_directory / 'no_context.template'),
        'target': str(target.name),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=Path(tmpdir),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    return_target = list(compile_action.execute().values())[0]

    assert return_target == target
    assert target.read_text() == 'one\ntwo\nthree'


def test_compilation_with_context(template_directory):
    """
    Templates should be compiled with the context store.

    It should compile differently after mutatinig the store.
    """
    compile_dict = {
        'content': 'test_template.conf',
    }
    context_store = {}

    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store=context_store,
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    context_store['fonts'] = {2: 'ComicSans'}
    target = list(compile_action.execute().values())[0]

    username = os.environ.get('USER')
    assert target.read_text() == f'some text\n{username}\nComicSans'

    context_store['fonts'] = {2: 'TimesNewRoman'}
    target = list(compile_action.execute().values())[0]
    assert target.read_text() == f'some text\n{username}\nTimesNewRoman'


def test_setting_permissions_of_target_template(template_directory):
    """Template target permission bits should be settable."""
    compile_dict = {
        'content': 'empty.template',
        'permissions': '707',
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    target = list(compile_action.execute().values())[0]
    assert (target.stat().st_mode & 0o777) == 0o707


def test_use_of_replacer(template_directory, tmpdir):
    """All options should be run through the replacer."""
    compile_dict = {
        'content': 'template',
        'target': 'target',
        'permissions': 'permissions',
    }

    template = template_directory / 'no_context.template'
    target = Path(tmpdir) / 'target'

    def replacer(string: str) -> str:
        """Trivial replacer."""
        if string == 'template':
            return template.name
        elif string == 'target':
            return str(target)
        elif string == 'permissions':
            return '777'
        else:
            return string

    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=replacer,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    target = list(compile_action.execute().values())[0]
    assert target.read_text() == 'one\ntwo\nthree'
    assert (target.stat().st_mode & 0o777) == 0o777


def test_that_current_directory_is_set_correctly(template_directory, tmpdir):
    """Shell commmand filters should be run from `directory`."""
    compile_dict = {
        'content': str(
            template_directory / 'shell_filter_working_directory.template',
        ),
    }

    directory = Path(tmpdir)
    compile_action = CompileAction(
        options=compile_dict,
        directory=directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    target = list(compile_action.execute().values())[0]
    assert target.read_text() == tmpdir


def test_retrieving_all_compiled_templates(template_directory, tmpdir):
    """Compile actions should return all compiled templates."""
    target1, target2 = Path(tmpdir) / 'target.tmp', Path(tmpdir) / 'target2'
    targets = [target1, target2]
    template = Path('no_context.template')
    compile_dict = {
        'content': str(template),
        'target': '{target}',
    }

    # First replace {target} with target1, then with target2, by doing some
    # trickery with the replacer function.
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x.format(
            target=targets.pop(),
        ) if x == '{target}' else x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    assert compile_action.performed_compilations() == {}

    compile_action.execute()
    assert compile_action.performed_compilations() == {
        template_directory / template: {target2},
    }

    compile_action.execute()
    assert compile_action.performed_compilations() == {
        template_directory / template: {target1, target2},
    }


def test_contains_special_method(template_directory, tmpdir):
    """Compile actions should 'contain' its compiled template."""
    temp_dir = Path(tmpdir)
    compile_dict = {
        'content': 'empty.template',
        'permissions': '707',
        'target': str(temp_dir / 'target.tmp'),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    compile_action.execute()
    assert template_directory / 'empty.template' in compile_action
    assert Path('/no/template') not in compile_action


def test_contains_with_uncompiled_template(template_directory, tmpdir):
    """Compile action only contains *compiled* templates."""
    temp_dir = Path(tmpdir)
    compile_dict = {
        'content': 'empty.template',
        'permissions': '707',
        'target': str(temp_dir / 'target.tmp'),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    assert template_directory / 'empty.template' not in compile_action

    compile_action.execute()
    assert template_directory / 'empty.template' in compile_action


def test_compiling_entire_directory(test_config_directory, tmpdir):
    """All directory contents should be recursively compiled."""
    temp_dir = Path(tmpdir).resolve()
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'

    # TODO: Make this unecessary
    for file in templates.glob('**/*.tmp'):
        file.unlink()

    compile_dict = {
        'content': str(templates),
        'target': str(temp_dir),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    results = compile_action.execute()

    # Check if return content is correct, showing performed compilations
    assert templates / 'module.template' in results
    assert results[templates / 'module.template'] == \
        temp_dir / 'module.template'

    # Check if the templates actually have been compiled
    target_dir_content = list(temp_dir.iterdir())
    assert len(target_dir_content) == 6
    assert temp_dir / 'module.template' in target_dir_content
    assert (temp_dir / 'recursive' / 'empty.template').is_file()


def test_filtering_compiled_templates(test_config_directory, tmpdir):
    """Users should be able to restrict compilable templates."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'
    compile_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'include': r'.+\.template',
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    compile_action.execute()

    # We should have a total of two compiled files
    assert len(list(temp_dir.iterdir())) == 2
    assert len(list((temp_dir / 'recursive').iterdir())) == 1
    assert (temp_dir / 'module.template').is_file()
    assert (temp_dir / 'recursive' / 'empty.template').is_file()


def test_renaming_templates(test_config_directory, tmpdir):
    """Templates targets should be renameable with a capture group."""
    temp_dir = Path(tmpdir)
    templates = \
        test_config_directory / 'test_modules' / 'using_all_actions'

    # Multiple capture groups should be allowed
    compile_dict = {
        'content': str(templates),
        'target': str(temp_dir),
        'include': r'(?:^template\.(.+)$|^(.+)\.template$)',
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store={'geography': {'capitol': 'Berlin'}},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    compile_action.execute()

    # We should have a total of two compiled files
    assert len(list(temp_dir.iterdir())) == 2
    assert len(list((temp_dir / 'recursive').iterdir())) == 1
    assert (temp_dir / 'module').is_file()
    assert (temp_dir / 'recursive' / 'empty').is_file()


def test_that_temporary_compile_targets_have_deterministic_paths(tmpdir):
    """Created compilation targets should be deterministic."""
    template_source = Path(tmpdir, 'template.tmp')
    template_source.write_text('content')

    compile_dict = {
        'content': str(template_source),
    }
    compile_action1 = CompileAction(
        options=compile_dict.copy(),
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    compile_action2 = CompileAction(
        options=compile_dict.copy(),
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    target1 = compile_action1.execute()[template_source]
    target2 = compile_action2.execute()[template_source]
    assert target1 == target2


def test_creation_of_backup(create_temp_files):
    """Existing external files should be backed up."""
    target, template = create_temp_files(2)

    # This file is the original and should be backed up
    target.write_text('original')

    # This is the new content compiled to target
    template.write_text('new')

    compile_dict = {
        'content': str(template.name),
        'target': str(target),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # We replace the content by executing the action
    compile_action.execute()
    assert target.read_text() == 'new'

    # And when cleaning up the module, the backup should be restored
    CreatedFiles().cleanup(module='test')
    assert target.read_text() == 'original'


def test_cleanup_of_created_directories(create_temp_files, tmpdir):
    """Directories created during compilation should be cleaned up."""
    tmpdir = Path(tmpdir)
    [template] = create_temp_files(1)

    # The target requires two new directories to be created
    directory = tmpdir / 'subdir' / 'dir'
    target = directory / 'target.tmp'

    compile_dict = {
        'content': str(template.name),
        'target': str(target),
    }
    compile_action = CompileAction(
        options=compile_dict,
        directory=template.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # At first, the target and its two parent directories do not exist
    assert not target.exists()
    assert not directory.exists()
    assert not directory.parent.exists()

    # Then, these three paths are created
    compile_action.execute()
    assert target.exists()
    assert directory.exists()
    assert directory.parent.exists()

    # And then must be cleaned up again
    CreatedFiles().cleanup(module='test')
    assert not target.exists()
    assert not directory.exists()
    assert not directory.parent.exists()
